"""Routines and class for Bogoliubov IMSRG diagrams."""

from future import standard_library
standard_library.install_aliases()
from builtins import range
import math
import networkx as nx
import numpy as np
from itertools import permutations, product
from more_itertools import powerset
import adg.diag
from adg.tools import reversed_enumerate


def two_partitions(number):
    """Return 2-partitions of the given integer.

    Args:
        numbr (int): The integer to partition.

    Returns:
        (list): All the 2-partitions as tuples.

    >>> two_partitions(3)
    [(0, 3), (1, 2), (2, 1), (3, 0)]

    >>> two_partitions(0)
    [(0, 0)]

    >>> two_partitions(-1)
    []

    """
    return [(elem, number - elem) for elem in range(number+1)]


def diagrams_generation(orders):
    """Generate diagrams for B-IMSRG.

    Args:
        orders (tuple): The B-IMSRG (N_A, N_B, N_C) order of the diagrams.

    Returns:
        (tuple): NumPy arrays encoding the adjacency matrices of the graphs,
            and number of diagrams with the A vertex on top.

    """
    deg_ext_max = 2*orders[-1]
    diagrams_topA = diagrams_subset(2*orders[0], 2*orders[1], deg_ext_max)
    diagrams_topB = diagrams_subset(2*orders[1], 2*orders[0], deg_ext_max)

    return diagrams_topA + diagrams_topB, len(diagrams_topA)


def diagrams_subset(deg_max_top, deg_max_bot, deg_max_ext):
    """Generate diagrams for B-IMSRG.

    Args:
        orders (tuple): The max ranks (2*N_A, 2*N_B, 2*N_C) of the vertices.

    Returns:
        (list): NumPy arrays encoding the adjacency matrices of the graphs.

    >>> diagrams_subset(2, 2, 0) # doctest: +NORMALIZE_WHITESPACE
    [array([[0, 0, 0, 0],
           [0, 0, 2, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 0]])]
    >>> len(diagrams_subset(2, 2, 2))
    5
    >>> len(diagrams_subset(4, 4, 4))
    41
    >>> len(diagrams_subset(0, 0, 0))
    0

    """
    matrices = []
    mat = np.zeros((4, 4), dtype=int)

    # Pick a valid vertex degree or zero for the external lines
    for deg_ext in range(0, deg_max_ext + 1, 2):
        # Split the valid vertex degree between external vertices
        for deg_0, deg_3 in two_partitions(deg_ext):
            # Split the vertex degree between the two internal vertices
            for part_0, part_3 in product(two_partitions(deg_0),
                                          two_partitions(deg_3)):
                mat[0,1], mat[0,2] = part_0
                mat[1,3], mat[2,3] = part_3
                temp_deg_1 = mat[0,1] + mat[1,3]
                temp_deg_2 = mat[0,2] + mat[2,3]
                # Check that internal vertices are both odd/even, as lines
                # between them will affect their degrees similarly
                if ((abs(temp_deg_1 - temp_deg_2) % 2) == 0):
                    # Determine how many lines can connect the vertices
                    max_addition = min(deg_max_bot - temp_deg_1,
                                       deg_max_top - temp_deg_2)
                    # Check the odd/even character of vertices, since they
                    # must eventually be even
                    min_addition = 2 if (temp_deg_1 % 2 == 0) else 1
                    for addition in range(min_addition, max_addition+1, 2):
                        temp_mat = mat.copy()
                        temp_mat[1,2] = addition
                        matrices.append(temp_mat)
    return matrices


def order_diagrams(diagrams, order):
    """Order the BIMSRG diagrams and return the number of diags for each type.

    Args:
        diagrams (list): The unordered BimsrgDiagrams.
        order (int): The order of the B-IMSRG truncation.

    Returns:
        (tuple): First element are the ordered BimsrgDiagrams. Second element
        is the number of diagrams for each type. Third element is flags for the
        output processing.

    """
    diags_per_order = {n: [] for n in range(1, order + 1)}

    for i_diag, diag in reversed_enumerate(diagrams):
        diags_per_order[diag.max_degree/2].append(diag)
        del diagrams[i_diag]

    diags_nb_per_type = {n: len(diags_per_order[n])
                         for n in range(1, order + 1)}

    diagrams = []
    for n in range(1, order + 1):
        diagrams += sorted(diags_per_order[n],
                           key=lambda diag: (diag.ext_io_degree,
                                             not diag.is_AB))
    diags_nb_per_type['nb_diags'] = len(diagrams)

    section_flags = {1: 0}

    for ind, diagram in enumerate(diagrams):
        diagram.tags[0] = ind
        if ind == 0:
            section_flags['new_op_struct'] = [ind]
        elif diagram.ext_io_degree != diagrams[ind-1].ext_io_degree:
            section_flags['new_op_struct'].append(ind)

    for n in range(2, order + 1):
        index = sum(len(diags_per_order[i]) for i in range(1, n))
        section_flags[n] = diagrams[index].tags[0] if diags_per_order[n] else -1

    return diagrams, diags_nb_per_type, section_flags


def write_header(tex_file, commands, diags_nbs):
    """Write overall header for BIMSRG result file.

    Args:
        tex_file (file): The ouput LaTeX file of the program.
        commands (Namespace): Flags for the program run.
        diags_nbs (dict): The number of diagrams per type.

    """
    tex_file.write("$C=\\left[A,B\\right]$ "
                   + "with $N_A = %i$, $N_B = %i$ and $N_C = %i$\n\n"
                   % commands.order)
    tex_file.write("$d_\\mathrm{max} \\equiv \\mathrm{max}(d_A, d_B, d_C)\n\n")
    tex_file.write("\\par \\vspace{\\baselineskip}\n")
    tex_file.write("Valid diagrams: %i\n\n" % diags_nbs['nb_diags'])

    for n in range(1, commands.order[-1] + 1):
        tex_file.write("$d_\\mathrm{max} = %i$ diagrams: %i\n\n"
                       % (n, diags_nbs[n]))


def write_permutator_section(tex_file, commands):
    """Write the section defining permutation operators.

    Args:
        tex_file (file): The ouput LaTeX file of the program.
        commands (Namespace): Flags for the program run.

    """
    tex_file.write("\\section{Permutators definitions}\n\n"
                   + "\\begin{align*}\n")
    # Number of external legs is at most two times N_C
    for max_index in range(2, 2*commands.order[-1] + 1):
        indices = list(range(1, max_index + 1))
        # Split our qp states in two subgroups of varying length
        for partition in ((indices[:n], indices[n:])
                          for n in range(1, len(indices)//2 + 1)):
            tex_file.write(permutator(partition[0], partition[1]))
    tex_file.write("\\end{align*}\n")


def permutator(set_1, set_2):
    """Write the definition of the permutation operator given by the two sets.

    Args:
        set_1 (list): The list of the left-hand-side qp labels.
        set_2 (list): The list of the right-hand-side qp labels.

    Returns:
        (str): The LaTeX expression for the permutation operator.

    >>> print(permutator([1, 2], [3])) # doctest: +NORMALIZE_WHITESPACE
    P(k_{1}k_{2}/k_{3}) &= 1 - P_{k_{1} k_{3}} - P_{k_{2} k_{3}} \\\\\n

    """
    perm_op = 'P(%s/%s) &= 1 ' % ("".join('k_{%i}' % label for label in set_1),
                                  "".join('k_{%i}' % label for label in set_2))
    perm_counter = 0
    # Combine all possible subsets of both sets
    for subset_1, subset_2 in product(powerset(set_1), powerset(set_2)):
        # Ensure each qp state has a permutation partner, exclude the empty set
        if (len(subset_1) == len(subset_2)) and (len(subset_1) > 0):
            # Run through all permutations of the first subset
            for permutation in permutations(subset_1):
                perm_op += '+ ' if len(subset_1) % 2 == 0 else '- '
                for elem in zip(permutation, subset_2):
                    perm_op += 'P_{k_{%i} k_{%i}} ' % elem
                    perm_counter += 1
                # Avoid permutators lines being too long
                if perm_counter >= 10:
                    perm_op += '\\\\\n &\\phantom{=} '
                    perm_counter = 0
    if perm_op.endswith('\\\\\n &\\phantom{=} '):
        perm_op = perm_op[:-len(' &\\phantom{=} ')]
    else:
        perm_op += "\\\\\n"
    return perm_op


class BimsrgDiagram(adg.diag.Diagram):
    """Describes a B-IMSRG Feynman diagram with its related properties.

    Attributes:
        adjacency_mat (Numpy array): The adjacency matrix of the diagram.
        unique_id (int): A unique number associated to the diagram.
        expr (str): The B-IMSRG expression associated to the diagram.
        ext_io_degree (tuple): The degree of the operator component the diagram
            corresponds to.
        is_AB (bool): True if the diagram contributes to +AB, false if to -BA.

    """

    __slots__ = ('adjacency_mat', 'unique_id', 'ext_io_degree',
                 '_vert_exchange_sym_fact', 'expr', 'is_AB')

    def __init__(self, nx_graph, tag_num):
        """Generate a BMBPT diagrams using a NetworkX graph.

        Args:
            nx_graph (NetworkX MultiDiGraph): The graph of interest.
            tag_num (int): The tag number associated to the diagram.

        """
        adg.diag.Diagram.__init__(self, nx_graph)
        self.tags = [tag_num]
        self.unique_id = tag_num
        self.adjacency_mat = nx.to_numpy_array(self.graph, dtype=int)
        self.expr = self.attribute_expression()
        self.max_degree = max(self.unsort_degrees[0] + self.unsort_degrees[3],
                              self.unsort_degrees[1],
                              self.unsort_degrees[2])
        self.ext_io_degree = (self.unsort_degrees[0], self.unsort_degrees[3])
        self.is_AB = True if self.graph.nodes[2]['operator'] == 'A' else False

    def attribute_expression(self):
        """Returns the LaTeX expression of the diagram.

        The expression is extracted in a way that assumes the canonical
        representation of the vertices as well as a labeling of the external
        lines that would correspond to a canonical representation of the C
        operator vertex. As such, there is no sign factor associated to
        departing from the canonical representation of the diagram. This
        additionally prevents any line crossing from appearing.

        Returns:
            (str): The LaTeX expression for the diagram.

        """
        A_degrees = self.unsort_io_degrees[1] \
            if self.graph.nodes[1]['operator'] == 'A' \
            else self.unsort_io_degrees[2]
        B_degrees = self.unsort_io_degrees[1] \
            if self.graph.nodes[1]['operator'] == 'B' \
            else self.unsort_io_degrees[2]
        name = " C^{%i%i}_{%s}(%i%i,%i%i) = " \
            % (self.unsort_degrees[3], self.unsort_degrees[0],
               "".join('k_{%i}' % (idx + 1) for idx
                       in range(self.unsort_degrees[3]
                                + self.unsort_degrees[0])),
               A_degrees[1], A_degrees[0], B_degrees[1], B_degrees[0])
        return name + self.sign() + self.permutator() \
            + self.symmetry_factor() + self.vertices_expression()

    def sign(self):
        """Return the sign of the diagram.

        As the diagrams are made with vertices in canonical representation and
        avoiding crossings, the only sign left is associated with the
        commutator.

        Returns:
            (str): The sign of the diagram.

        """
        return '- ' if self.graph.nodes[2]['operator'] == 'B' else ''

    def permutator(self):
        """Return the permutator associated to the diagram.

        The labelling of the external lines correspond to the canonical
        representation of the C operator vertex.

        Returns:
            (str): The permutator associated to the diagram in LaTeX format.

        """
        # Number of external lines tied to each vertex
        nb_out_2 = self.adjacency_mat[2, 3]
        nb_in_2 = self.adjacency_mat[0, 2]
        nb_out_1 = self.adjacency_mat[1, 3]
        nb_in_1 = self.adjacency_mat[0, 1]

        perm = ""

        # Permutator for out-going lines
        if (nb_out_2 != 0) and (nb_out_1 != 0):
            perm += "P(%s/%s) " \
                % ("".join('k_{%i}' % (idx + 1) for idx in range(nb_out_2)),
                   "".join('k_{%i}' % (idx + 1) for idx
                           in range(nb_out_2, nb_out_2 + nb_out_1)))

        # Permutator for incoming lines
        if (nb_in_2 != 0) and (nb_in_1 != 0):
            perm += "P(%s/%s) " \
                % ("".join('k_{%i}' % (idx + 1) for idx
                           in range(nb_out_2 + nb_out_1,
                                    nb_out_2 + nb_out_1 + nb_in_2)),
                   "".join('k_{%i}' % (idx + 1) for idx
                           in range(nb_out_2 + nb_out_1 + nb_in_2,
                                    nb_out_2 + nb_out_1 + nb_in_2 + nb_in_1)))
        return perm

    def symmetry_factor(self):
        """Returns the symmetry factor of the diagram in LaTeX format.

        Returns:
            (str): The LaTeX-formatted symmetry factor.

        """
        # Count the number of internal lines
        factor = math.factorial(self.adjacency_mat[1, 2])
        return "\\frac{1}{%i}" % factor if factor != 1 else ""

    def vertices_expression(self):
        """Return the expression associated to the vertices in LaTeX format.

        The expressions are extracted in a way that assumes the canonical
        representation of the vertices as well as a labeling of the external
        lines that would correspond to a canonical representation of the C
        operator vertex. This prevents any crossing as well as additional sign
        tied to departing from the canonical representation.

        Returns:
            (str): The LaTeX-formatted expression for the vertices.
        """
        internal_lines = "".join("p_{%i}" % label for label
                                 in range(1, self.adjacency_mat[1, 2] + 1))
        # Number of external lines tied to each vertex
        nb_out_2 = self.adjacency_mat[2, 3]
        nb_in_2 = self.adjacency_mat[0, 2]
        nb_out_1 = self.adjacency_mat[1, 3]
        nb_in_1 = self.adjacency_mat[0, 1]
        # Expression associated to the upper vertex
        expr_2 = "%s^{%i%i}_{%s %s}" \
            % (self.graph.nodes[2]['operator'],
               self.unsort_io_degrees[2][1], self.unsort_io_degrees[2][0],
               "".join('k_{%i}' % (idx + 1) for idx in range(nb_out_2))
               + "".join('k_{%i}' % (idx + 1) for idx
                         in range(nb_out_2 + nb_out_1,
                                  nb_out_2 + nb_out_1 + nb_in_2)),
               internal_lines)
        # Expression associated to the lower vertex
        expr_1 = "%s^{%i%i}_{%s %s}" \
            % (self.graph.nodes[1]['operator'],
               self.unsort_io_degrees[1][1], self.unsort_io_degrees[1][0],
               internal_lines,
               "".join('k_{%i}' % (idx + 1) for idx
                       in range(nb_out_2, nb_out_2 + nb_out_1))
               + "".join('k_{%i}' % (idx + 1) for idx
                         in range(nb_out_2 + nb_out_1 + nb_in_2,
                                  nb_out_2 + nb_out_1 + nb_in_2 + nb_in_1)))

        return "\\sum_{%s} %s %s" % (internal_lines, expr_2, expr_1)

    def write_section(self, result, commands, section_flags):
        """Write section and subsections for BMBPT result file.

        Args:
            result (file): The LaTeX output file of the program.
            commands (dict): The flags associated with run management.
            section_flags (dict): UniqueIDs of diags starting each section.

        """
        for n in range(1, commands.order[-1] + 1):
            if self.tags[0] == section_flags[n]:
                result.write("\\section{$d_\\mathrm{max} = %i$}\n\n" % n)
        if self.tags[0] in section_flags['new_op_struct']:
            result.write("\\subsection{$C^{%i%i}$}\n\n"
                         % (self.ext_io_degree[1], self.ext_io_degree[0]))
        result.write("\\paragraph{Diagram %i (%s):}\n"
                     % (self.tags[0] + 1, '$+AB$' if self.is_AB else '$-BA$'))
        result.write("\\begin{equation}\n%s\n\\end{equation}\n" % self.expr)



"""Routines and class for all types of diagrams, inherited by others."""

from builtins import range
from builtins import object, str
from adg.tools import reversed_enumerate

import copy
import numpy
import networkx as nx


def no_trace(matrices):
    """Select matrices with full 0 diagonal.

    Args:
        matrices (list): A list of adjacency matrices.

    Returns:
        (list): The adjacency matrices without non-zero diagonal elements.

    >>> test_matrices = [[[0, 1, 2], [2, 0, 1], [5, 2, 0]], \
    [[2, 2, 2], [1, 2, 3], [0, 0, 0]], \
    [[0, 1, 3], [2, 0, 8], [2, 1, 0]]]
    >>> no_trace(test_matrices)
    [[[0, 1, 2], [2, 0, 1], [5, 2, 0]], [[0, 1, 3], [2, 0, 8], [2, 1, 0]]]

    """
    traceless_matrices = []
    for matrix in matrices:
        for ind_i, line in enumerate(matrix):
            if line[ind_i] != 0:
                break
        else:
            traceless_matrices.append(matrix)
    return traceless_matrices


def check_vertex_degree(matrices, three_body_use, nbody_max_observable,
                        canonical_only, vertex_id):
    """Check the degree of a specific vertex in a set of matrices.

    Args:
        matrices (list): Adjacency matrices.
        three_body_use (bool): ``True`` if one uses three-body forces.
        nbody_max_observable (int): Maximum body number for the observable.
        canonical_only (bool): ``True`` if one draws only canonical diagrams.
        vertex_id (int): The position of the studied vertex.

    >>> test_matrices = [numpy.array([[0, 1, 2], [1, 0, 1], [0, 2, 0]]), \
        numpy.array([[2, 0, 2], [1, 2, 3], [1, 0, 0]]), \
        numpy.array([[0, 1, 3], [2, 0, 8], [2, 1, 0]])]
    >>> check_vertex_degree(test_matrices, True, 3, False, 0)
    >>> test_matrices #doctest: +NORMALIZE_WHITESPACE
    [array([[0, 1, 2], [1, 0, 1], [0, 2, 0]]),
     array([[2, 0, 2], [1, 2, 3], [1, 0, 0]])]
    >>> check_vertex_degree(test_matrices, False, 2, False, 0)
    >>> test_matrices #doctest: +NORMALIZE_WHITESPACE
    [array([[0, 1, 2], [1, 0, 1], [0, 2, 0]])]

    """
    authorized_deg = [4]
    if three_body_use:
        authorized_deg.append(6)
    if not canonical_only or vertex_id == 0:
        authorized_deg.append(2)
    authorized_deg = tuple(authorized_deg)

    for i_mat, matrix in reversed_enumerate(matrices):
        vertex_degree = sum(matrix[index][vertex_id] + matrix[vertex_id][index]
                            for index in list(range(matrix.shape[0])))
        vertex_degree -= matrix[vertex_id][vertex_id]

        if (vertex_id != 0 and vertex_degree not in authorized_deg) \
                or (vertex_id == 0 and vertex_degree > 2*nbody_max_observable):
            del matrices[i_mat]


def topologically_distinct_diagrams(diagrams):
    """Return a list of diagrams all topologically distinct.

    Args:
        diagrams (list): The Diagrams of interest.

    Returns:
        (list): Topologically unique diagrams.

    """
    import adg.tsd
    iso = nx.algorithms.isomorphism
    op_nm = iso.categorical_node_match('operator', False)
    anom_em = iso.categorical_multiedge_match('anomalous', False)
    for i_diag, diag in reversed_enumerate(diagrams):
        graph = diag.graph
        diag_io_degrees = diag.io_degrees
        for i_comp_diag, comp_diag in reversed_enumerate(diagrams[:i_diag]):
            if diag_io_degrees == comp_diag.io_degrees:
                # Check anomalous character of props for PBMBPT
                if isinstance(diag, adg.pbmbpt.ProjectedBmbptDiagram):
                    doubled_graph = create_checkable_diagram(graph)
                    doubled_comp_diag = create_checkable_diagram(comp_diag.graph)
                    matcher = iso.DiGraphMatcher(doubled_graph,
                                                 doubled_comp_diag,
                                                 node_match=op_nm,
                                                 edge_match=anom_em)
                # Check for topologically equivalent diags considering vertex
                # properties but not edge attributes, i.e. anomalous character
                else:
                    matcher = iso.DiGraphMatcher(graph,
                                                 comp_diag.graph,
                                                 node_match=op_nm)
                if matcher.is_isomorphic():
                    # Store the set of permutations to recover the original TSD
                    if isinstance(diag, adg.tsd.TimeStructureDiagram):
                        diag.perms.update(
                            update_permutations(comp_diag.perms,
                                                comp_diag.tags[0],
                                                matcher.mapping)
                            )
                    diag.tags += comp_diag.tags
                    del diagrams[i_comp_diag]
                    break
    return diagrams


def update_permutations(comp_graph_perms, comp_graph_tag, mapping):
    """Update permutations associated to the BMBPT diags for a shared TSD.

    Args:
        comp_graph_perms (dict): Permutations to be updated.
        comp_graph_tag (int): The tag associated to the TSD configuration.
        mapping (dict): permutations to go from previous ref TSD to new one.

    """
    identity = {key: key for key in comp_graph_perms[comp_graph_tag]}
    # Do permutations only when necessary
    if mapping != identity:
        for graph_id in comp_graph_perms:
            # Create a dummy dictionary to avoid overwriting some nodes
            dummy_nodes = copy.deepcopy(comp_graph_perms[graph_id])
            # Permute the nodes according to the new mapping
            for node in comp_graph_perms[graph_id]:
                comp_graph_perms[graph_id][node] = dummy_nodes[mapping[node]]

    return comp_graph_perms


def create_checkable_diagram(pbmbpt_graph):
    """Return a graph with anomalous props going both ways for topo check.

    Args:
        pbmbpt_graph (NetworkX MultiDiGraph): The graph to be copied.

    Returns:
        (NetworkX MultiDiGraph): Graph with double the anomalous props.

    """
    doubled_graph = copy.deepcopy(pbmbpt_graph)
    props_to_add = [(prop[1], prop[0]) for prop
                    in doubled_graph.edges(keys=True, data='anomalous')
                    if prop[3] and not prop[0] == prop[1]]
    doubled_graph.add_edges_from(props_to_add, anomalous=True, weight=1)
    return doubled_graph


def label_vertices(graphs_list, theory_type, switch_flag):
    """Account for different status of vertices in operator diagrams.

    Args:
        graphs_list (list): The Diagrams of interest.
        theory_type (str): The name of the theory of interest.
        switch_flag (int): When to switch A and B operators for BIMSRG.

    """
    if theory_type != 'BIMSRG':
        for graph in graphs_list:
            nx.set_node_attributes(graph, False, 'operator')
            if theory_type in ("BMBPT", "PBMBPT"):
                graph.nodes[0]['operator'] = True
    else:
        for idx, graph in enumerate(graphs_list):
            nx.set_node_attributes(graph, '', 'operator')
            if idx < switch_flag:
                graph.nodes[1]['operator'] = 'B'
                graph.nodes[2]['operator'] = 'A'
            else:
                graph.nodes[1]['operator'] = 'A'
                graph.nodes[2]['operator'] = 'B'


def feynmf_generator(graph, theory_type, diagram_name):
    """Generate the feynmanmp instructions corresponding to the diagram.

    Args:
        graph (NetworkX MultiDiGraph): The graph of interest.
        theory_type (str): The name of the theory of interest.
        diagram_name (str): The name of the studied diagram.

    """
    p_order = graph.number_of_nodes()
    diag_size = 20*p_order

    theories = ["MBPT", "BMBPT", "PBMBPT", "BIMSRG"]
    prop_types = ["half_prop", "prop_pm", "prop_pm", "half_prop"]
    propa = prop_types[theories.index(theory_type)]

    fmf_file = open(diagram_name + ".tex", 'w')
    fmf_file.write("\\parbox{40pt}{\\begin{fmffile}{%s}\n" % diagram_name
                   + "\\begin{fmfgraph*}(40,%i)\n" % diag_size)

    # Define the appropriate line propagator_style
    fmf_file.write(propagator_style(propa))
    if theory_type == "PBMBPT":
        fmf_file.write(propagator_style("prop_mm"))

    # Set the position of the vertices
    if theory_type == "BIMSRG":
        bimsrg_diagram_internals(graph, fmf_file, propa)

    else:
        fmf_file.write(vertex_positions(graph, p_order))

        # Special config for self-contraction
        if theory_type == "PBMBPT":
            fmf_file.write(self_contractions(graph))

        # Loop over all elements of the graph to draw associated propagators
        for vert_i in graph:
            for vert_j in list(graph.nodes())[vert_i+1:]:
                props_to_draw = [edge for edge in graph.edges([vert_i, vert_j],
                                                              data=True,
                                                              keys=True)
                                 if edge[1] in (vert_i, vert_j)
                                 and edge[0] != edge[1]]
                # Set the list of propagators directions to use
                props_dir = prop_directions(vert_j - vert_i, len(props_to_draw))
                # Draw the diagrams
                key = 0
                # Start with props going down, used in MBPT only
                for _ in (p for p in props_to_draw
                          if p[1] < p[0]
                          and not ('anomalous' in p[3] and p[3]['anomalous'])):
                    fmf_file.write("\\fmf{%s%s}{v%i,v%i}\n"
                                   % (propa, props_dir[key], vert_j, vert_i))
                    key += 1
                # Reinitialise the drawing configuration as we change direction
                key = 0
                for _ in (p for p in props_to_draw
                          if ('anomalous' in p[3] and p[3]['anomalous'])):
                    fmf_file.write("\\fmf{prop_mm%s}{v%i,v%i}\n"
                                   % (props_dir[key], vert_i, vert_j))
                    key += 1
                for _ in (p for p in props_to_draw
                          if p[0] < p[1]
                          and not ('anomalous' in p[3] and p[3]['anomalous'])):
                    fmf_file.write("\\fmf{%s%s}{v%i,v%i}\n"
                                   % (propa, props_dir[key], vert_i, vert_j))
                    key += 1

    fmf_file.write("\\end{fmfgraph*}\n\\end{fmffile}}\n")
    fmf_file.close()


def prop_directions(vert_distance, nb_props):
    """Return a list of possible propagators directions.

    Args:
        vert_distance (int): Distance between the two connected vertices.
        nb_props (int): Number of propagators to be drawn.

    Returns:
        (list): Propagators directions stored as strings.

    """
    if nb_props < 7:
        directions = [",right=0.9", ",right=0.75", ",right=0.6", ",right=0.5",
                      "", ",left=0.5", ",left=0.6", ",left=0.75", ",left=0.9"]

        if vert_distance != 1:
            props_dir = directions[:3] + directions[-3:]
        else:
            props_dir = directions[:2] + directions[3:6] + directions[-2:]
            if nb_props % 2 != 1:
                props_dir = props_dir[:3] + props_dir[-3:]
            else:
                props_dir = props_dir[1:]
        if nb_props < 5:
            props_dir = props_dir[1:-1]
            if nb_props < 3:
                props_dir = props_dir[1:-1]

    elif vert_distance == 1:
        directions = [",right=0.%i" % angle for angle in range(90, 0, -10)] \
            + [",left=0.%i" % angle for angle in range(10, 100, 10)]

        props_dir = directions[nb_props//4:-(nb_props//4)]

    return props_dir


def propagator_style(prop_type):
    """Return the FeynMF definition for the appropriate propagator type.

    Args:
        prop_type (str): The type of propagators used in the diagram.

    Returns:
        (str): The FeynMF definition for the propagator style used.

    """
    line_styles = {}

    line_styles['prop_pm'] = "\\fmfcmd{style_def prop_pm expr p =\n" \
        + "draw_plain p;\nshrink(.7);\n" \
        + "\tcfill (marrow (p, .25));\n" \
        + "\tcfill (marrow (p, .75))\n" \
        + "endshrink;\nenddef;}\n"

    line_styles['prop_mm'] = "\\fmfcmd{style_def prop_mm expr p =\n" \
        + "draw_plain p;\nshrink(.7);\n" \
        + "\tcfill (marrow (p, .75));\n" \
        + "\tcfill (marrow (reverse p, .75))\n" \
        + "endshrink;\nenddef;}\n"

    line_styles['half_prop'] = "\\fmfcmd{style_def half_prop expr p =\n" \
        + "draw_plain p;\nshrink(.7);\n" \
        + "\tcfill (marrow (p, .5))\n" \
        + "endshrink;\nenddef;}\n"

    return line_styles[prop_type]


def vertex_positions(graph, order):
    """Return the positions of the graph's vertices.

    Args:
        graph (NetworkX MultiDiGraph): The graph of interest.
        order (int): The perturbative order of the graph.

    Returns:
        (str): The FeynMP instructions for positioning the vertices.

    """
    positions = "\\fmftop{v%i}\\fmfbottom{v0}\n" % (order-1)
    for vert in range(order-1):
        positions += "\\fmf{phantom}{v%i,v%i}\n" % (vert, (vert+1)) \
            + ("\\fmfv{d.shape=square,d.filled=full,d.size=3thick"
               if 'operator' in graph.nodes[vert] and graph.nodes[vert]['operator']
               else "\\fmfv{d.shape=circle,d.filled=full,d.size=3thick") \
            + "}{v%i}\n" % vert
    positions += "\\fmfv{d.shape=circle,d.filled=full,d.size=3thick}{v%i}\n" \
        % (order-1) + "\\fmffreeze\n"
    return positions


def self_contractions(graph):
    """Return the instructions for drawing the graph's self-contractions.

    Args:
        graph (NetworkX MultiDiGraph): The graph being drawn.

    Returns:
        (str): FeynMF instructions for drawing the self-contractions.

    """
    instructions = ""
    # Check for self-contractions before going further
    if [nx.selfloop_edges(graph)]:
        instructions += propagator_style("half_prop")
        for vert in graph:
            props_to_draw = [edge for edge
                             in nx.selfloop_edges(graph, data=True, keys=True)
                             if edge[0] == vert]
            positions = ["15pt", "-15pt"]
            key = 0
            for prop in props_to_draw:
                if prop[3]['anomalous']:
                    a_name = "a%i%i" % (vert, key)
                    instructions += ("\\fmfv{}{%s}\n" % a_name
                                     + "\\fmffixed{(%s,0)}{v%i,%s}\n"
                                     % (positions[key], vert, a_name)
                                     + "\\fmf{half_prop,right}{%s,v%i}\n"
                                     % (a_name, vert)
                                     + "\\fmf{half_prop,left}{%s,v%i}\n"
                                     % (a_name, vert))
                    key += 1
        instructions += "\\fmffreeze\n"
    return instructions


def bimsrg_diagram_internals(graph, fmf_file, prop_type):
    """Write to file the vertices and propagators of BIMSRG diagrams.

    Args:
        graph (NetworkX MultiDiGraph): The graph to be drawn.
        fmf_file (file): The FeymanMF file to be written to.
        prop_type (str): The FeymanMF type for drawing the propagators.

    """
    nbs_out_edges = (sum(1 for p in graph.in_edges(3, keys=True) if p[0] == 1),
                     sum(1 for p in graph.in_edges(3, keys=True) if p[0] == 2))
    nbs_in_edges = (sum(1 for p in graph.out_edges(0, keys=True) if p[1] == 1),
                    sum(1 for p in graph.out_edges(0, keys=True) if p[1] == 2))

    nb_out_edges = len(graph.in_edges(3, keys=True))
    nb_in_edges = len(graph.out_edges(0, keys=True))

    nb_top_vertices = nb_out_edges if nb_out_edges == 1 \
        else (max(2*nbs_out_edges[0], 2*nbs_out_edges[1]) + 1)
    nb_bot_vertices = nb_in_edges if nb_in_edges == 1 \
        else (max(2*nbs_in_edges[0], 2*nbs_in_edges[1]) + 1)

    positions = "\\fmfstraight\n" \
        + "\\fmftopn{t}{%i}" % nb_top_vertices \
        + "\\fmfbottomn{b}{%i}\n" % nb_bot_vertices \
        + "\\fmf{phantom}{b%i,v1}\n" % (nb_bot_vertices//2 + 1) \
        + "\\fmf{phantom}{v1,v2}\n" \
        + "\\fmf{phantom}{v2,t%i}\n" % (nb_top_vertices//2 + 1) \
        + "\\fmfv{d.shape=circle,d.filled=%s,d.size=3thick}{v1}\n" \
        % ('full' if graph.nodes[1]['operator'] == 'A' else 'empty') \
        + "\\fmfv{d.shape=circle,d.filled=%s,d.size=3thick}{v2}\n" \
        % ('full' if graph.nodes[2]['operator'] == 'A' else 'empty') \
        + "\\fmffreeze\n"
    fmf_file.write(positions)

    # Internal lines
    nb_props = sum(1 for edge in graph.edges(1, keys=True) if edge[1] == 2)
    # Set the list of propagators directions to use
    props_dir = prop_directions(1, nb_props)
    # Draw the propagators
    for idx in range(nb_props):
        fmf_file.write("\\fmf{%s%s}{v1,v2}\n" % (prop_type, props_dir[idx]))

    # Incoming external line
    for vertex in (1, 2):
        nb_props = sum(1 for edge in graph.edges(0, keys=True)
                       if edge[1] == vertex)
        if (nb_bot_vertices == 1) and (vertex == 1):
            orientation = ""
        else:
            orientation = ",left=0.4" if vertex == 2 else ",right=0.3"
        # Draw the propagators
        for key in range(nb_props):
            fmf_file.write("\\fmf{%s%s}{b%i,v%i}\n"
                           % (prop_type,
                              orientation,
                              key+1 if vertex == 2 else nb_bot_vertices - key,
                              vertex))

    # Outgoing external lines
    for vertex in (1, 2):
        nb_props = sum(1 for edge in graph.in_edges(3, keys=True)
                       if edge[0] == vertex)
        if (nb_top_vertices == 1) and (vertex == 2):
            orientation = ""
        else:
            orientation = ",right=0.4" if vertex == 1 else ",left=0.3"
        # Draw the propagators
        for key in range(nb_props):
            fmf_file.write("\\fmf{%s%s}{v%i,t%i}\n"
                           % (prop_type,
                              orientation,
                              vertex,
                              key+1 if vertex == 2 else nb_top_vertices - key))


def draw_diagram(directory, result_file, diagram_index, diag_type):
    """Copy the diagram feynmanmp instructions in the result file.

    Args:
        directory (str): The path to the output folder.
        result_file (file): The LaTeX ouput file of the program.
        diagram_index (str): The number associated to the diagram.
        diag_type (str): The type of diagram used here.

    """
    with open(directory+"/Diagrams/%s_%s.tex"
              % (diag_type, diagram_index)) as diag_file:
        result_file.write(diag_file.read())


def to_skeleton(graph):
    """Return the bare skeleton of a graph, i.e. only non-redundant links.

    Args:
        graph (NetworkX MultiDiGraph): The graph to be turned into a skeleton.

    Returns:
        (NetworkX MultiDiGraph): The skeleton of the initial graph.

    """
    for vertex_a in graph:
        for vertex_b in graph:
            while graph.number_of_edges(vertex_a, vertex_b) > 1:
                graph.remove_edge(vertex_a, vertex_b)
            if len(list(nx.all_simple_paths(graph, vertex_a, vertex_b))) > 1:
                while len(nx.shortest_path(graph, vertex_a, vertex_b)) == 2:
                    graph.remove_edge(vertex_a, vertex_b)
    return graph


def extract_denom(start_graph, subgraph):
    """Extract the appropriate denominator using the subgraph rule.

    Args:
        start_graph (NetworkX MultiDiGraph): The studied graph.
        subgraph (NetworkX MultiDiGraph): The subgraph used for this particular
            denominator factor.

    Returns:
        (str): The denominator factor for this subgraph.

    """
    denomin = r"\epsilon^{" \
        + "".join("%s"
                  % prop[3]['qp_state']
                  for prop
                  in start_graph.out_edges(subgraph, keys=True, data=True)
                  if not subgraph.has_edge(prop[0], prop[1], prop[2])
                  and not ('anomalous' in prop[3] and prop[3]['anomalous'])) \
        + "}_{" \
        + "".join("%s"
                  % prop[3]['qp_state']
                  for prop
                  in start_graph.in_edges(subgraph, keys=True, data=True)
                  if not subgraph.has_edge(prop[0], prop[1], prop[2])
                  and not ('anomalous' in prop[3] and prop[3]['anomalous'])) \
        + "".join("%s"
                  % prop[3]['qp_state']
                  for prop
                  in start_graph.in_edges(subgraph, keys=True, data=True)
                  if subgraph.has_edge(prop[0], prop[1], prop[2])
                  and ('anomalous' in prop[3] and prop[3]['anomalous'])) \
        + "".join("%s"
                  % (prop[3]['qp_state'].split("}")[0] + "}")
                  for prop
                  in start_graph.in_edges(subgraph, keys=True, data=True)
                  if not subgraph.has_edge(prop[0], prop[1], prop[2])
                  and ('anomalous' in prop[3] and prop[3]['anomalous'])) \
        + "".join("%s"
                  % (prop[3]['qp_state'].split("}")[1] + "}")
                  for prop
                  in start_graph.out_edges(subgraph, keys=True, data=True)
                  if not subgraph.has_edge(prop[0], prop[1], prop[2])
                  and ('anomalous' in prop[3] and prop[3]['anomalous'])) \
        + "}"
    return denomin


def print_adj_matrices(directory, diagrams):
    """Print a computer-readable file with the diagrams' adjacency matrices.

    Args:
        directory (str): The path to the output directory.
        diagrams (list): All the diagrams.

    """
    with open(directory+"/adjacency_matrices.txt", "w") as mat_file:
        for idx, diagram in enumerate(diagrams):
            mat_file.write("Diagram n: %i\n" % (idx + 1))
            numpy.savetxt(mat_file,
                          nx.to_numpy_array(diagram.graph, dtype=int),
                          fmt='%d')
            mat_file.write("\n")


class Diagram(object):
    """Describes a diagram with its related properties.

    Attributes:
        graph (NetworkX MultiDiGraph): The actual graph.
        unsorted_degrees (tuple): The degrees of the graph vertices
        degrees (tuple): The ascendingly sorted degrees of the graph vertices.
        unsort_io_degrees (tuple): The list of in- and out-degrees for each
            vertex of the graph, stored in a (in, out) tuple.
        io_degrees (tuple): The sorted version of unsort_io_degrees.
        max_degree (int): The maximal degree of a vertex in the graph.
        tags (list): The tag numbers associated to a diagram.

    """

    __slots__ = ('graph', 'unsort_degrees', 'degrees', 'unsort_io_degrees',
                 'io_degrees', 'max_degree', 'tags')

    def __init__(self, nx_graph):
        """Generate a Diagram object starting from the NetworkX graph.

        Args:
            nx_graph (NetworkX MultiDiGraph): The graph of interest.

        """
        self.graph = nx_graph
        self.unsort_degrees = tuple(nx_graph.degree(node) for node in nx_graph)
        self.degrees = tuple(sorted(self.unsort_degrees))
        self.unsort_io_degrees = tuple((nx_graph.in_degree(node),
                                        nx_graph.out_degree(node))
                                       for node in nx_graph)
        self.io_degrees = tuple(sorted(self.unsort_io_degrees))
        self.max_degree = self.degrees[-1]
        self.tags = [0]

    def write_graph(self, latex_file, directory, write_time):
        """Write the graph of the diagram to the LaTeX file.

        Args:
            latex_file (file): The LaTeX ouput file of the program.
            directory (str): Path to the result folder.
            write_time (bool): (Here to emulate polymorphism).

        """
        latex_file.write('\n\\begin{center}\n')
        draw_diagram(directory, latex_file, str(self.tags[0]), 'diag')
        latex_file.write('\n\\end{center}\n\n')



"""Routines and class for Many-Body Perturbation Theory diagrams."""
from __future__ import print_function
from __future__ import division

from builtins import map
from builtins import range
from adg.tools import reversed_enumerate
import copy
import itertools
import string
import numpy as np
import networkx as nx
import adg.diag


def diagrams_generation(order):
    """Generate the diagrams for the MBPT case.

    Args:
        order (int): The perturbative order of interest.

    Returns:
        (list): A list of NumPy arrays with the diagrams adjacency matrices.

    >>> diagrams_generation(2) # doctest: +NORMALIZE_WHITESPACE
    [array([[0, 2], [2, 0]])]
    >>> diagrams_generation(3) # doctest: +NORMALIZE_WHITESPACE
    [array([[0, 2, 0], [0, 0, 2], [2, 0, 0]]),
     array([[0, 1, 1], [1, 0, 1], [1, 1, 0]]),
     array([[0, 0, 2], [2, 0, 0], [0, 2, 0]])]
    >>> diagrams_generation(1)
    []

    """
    # Generate all 1-magic square of dimension order
    seeds = [k for k in itertools.permutations(list(range(order)), order)]
    all_matrices = [[[0 if i != j else 1 for i in range(order)]
                     for j in k]
                    for k in seeds]
    traceless = adg.diag.no_trace(all_matrices)
    coeffs = [i for i in itertools.combinations_with_replacement(
        list(range(len(traceless))), 2)]
    double = []

    for coef in coeffs:
        matrix = copy.deepcopy(traceless[coef[0]])
        for i, line in enumerate(traceless[coef[1]]):
            for j, elem in enumerate(line):
                matrix[i][j] += elem
        double.append(matrix)
    double_uniq = []
    for matrix in double:
        if matrix not in double_uniq:
            double_uniq.append(matrix)
    double_uniq.sort(reverse=True)
    return [np.array(matrix) for matrix in double_uniq]


def write_diag_exp(latex_file, mbpt_diag):
    """Write the expression associated to a diagram in the LaTeX file.

    Args:
        latex_file (file): The LaTeX output file to be written in.
        mbpt_diag (MbptDiagram): The diagram which expression is being written.

    """
    latex_file.write("\\begin{equation}\n")
    latex_file.write(mbpt_diag.expr)
    latex_file.write("\\end{equation}\n")


def write_header(tex_file, diags_nbs):
    """Write tha appropriate header for the LaTeX file for MBPT diagrams.

    Args:
        tex_file (file): The LaTeX ouput file to be written in.
        diags_nbs (dict): A dict with the number of diagrams per
            excitation level type.

    """
    tex_file.write("Valid diagrams: %i\n\n" % diags_nbs['nb_diags']
                   + "Singles: %i\n\n" % diags_nbs['singles']
                   + "Doubles: %i\n\n" % diags_nbs['doubles']
                   + "Triples: %i\n\n" % diags_nbs['triples']
                   + "Quadruples: %i\n\n" % diags_nbs['quadruples']
                   + "Quintuples and higher excitation levels: %i\n\n"
                   % diags_nbs['quintuples+'])


def print_cd_output(directory, diagrams):
    """Print a computer-readable file for automated frameworks.

    Args:
        directory (str): The path to the output directory.
        diagrams (list): All the MbptDiagrams.

    """
    cd_file = open(directory + '/CD_output.txt', 'w')
    conjug_file = open(directory + '/CD_conjug_pairs.list', 'w')
    for diag in diagrams:
        cd_file.write('config[%i] = %s\n' % (diag.tags[0] + 1, diag.cd_expr))
        if (diag.complex_conjugate != -1) \
                and (diag.complex_conjugate > diag.tags[0]):
            conjug_file.write("%i\t%i\n" % (diag.tags[0] + 1,
                                            diag.complex_conjugate + 1))
    cd_file.write('\n')
    cd_file.close()
    conjug_file.close()


def order_diagrams(diagrams):
    """Order the MBPT diagrams and return the number of diags for each type.

    Args:
        diagrams (list): The unordered MbptDiagrams.

    Returns:
        (tuple): First element are the ordered MbptDiagrams. Second element is
        the number of diagrams for each excitation level type.

    """
    singles = []
    doubles = []
    triples = []
    quadruples = []
    quintuples_and_higher = []

    for i_diag, diag in reversed_enumerate(diagrams):
        if diag.excitation_level == 1:
            singles.append(diag)
        elif diag.excitation_level == 2:
            doubles.append(diag)
        elif diag.excitation_level == 3:
            triples.append(diag)
        elif diag.excitation_level == 4:
            quadruples.append(diag)
        elif diag.excitation_level >= 5:
            quintuples_and_higher.append(diag)
        else:
            print("Zero or negative excitation level!\n")
            exit()
        del diagrams[i_diag]

    diagrams = singles + doubles + triples + quadruples \
        + quintuples_and_higher

    for ind, diagram in enumerate(diagrams):
        diagram.tags[0] = ind

    attribute_conjugate(diagrams)

    diags_nb_per_type = {
        'nb_diags': len(diagrams),
        'singles': len(singles),
        'doubles': len(doubles),
        'triples': len(triples),
        'quadruples': len(quadruples),
        'quintuples+': len(quintuples_and_higher)
        }

    section_flags = {
        'singles': singles[0].tags[0] if singles else -1,
        'doubles': doubles[0].tags[0] if doubles else -1,
        'triples': triples[0].tags[0] if triples else -1,
        'quadruples': quadruples[0].tags[0] if quadruples else -1,
        'quintuples+': quintuples_and_higher[0].tags[0]
                       if quintuples_and_higher else -1
    }

    return diagrams, diags_nb_per_type, section_flags


def attribute_conjugate(diagrams):
    """Attribute to each diagram its complex conjugate.

    The diagrams involved in conjugate pairs receive the tag associated to
    their partner in the ``complex_conjugate`` attribute.

    Args:
        diagrams (list): The topologically unique MbptDiagrams.

    """
    for idx, diag1 in enumerate(diagrams):
        if diag1.complex_conjugate == -1:
            for diag2 in diagrams[idx+1:]:
                if diag2.complex_conjugate == -1 \
                        and diag1.is_complex_conjug_of(diag2):
                    diag1.complex_conjugate = diag2.tags[0]
                    diag2.complex_conjugate = diag1.tags[0]
                    break


def extract_cd_denom(start_graph, subgraph):
    """Extract the computer-readable denominator using the subgraph rule.

    Args:
        start_graph (NetworkX MultiDiGraph): The studied graph.
        subgraph (NetworkX MultiDiGraph): The subgaph for this particular
            factor.

    Returns:
        (str): The denominator factor associated to this subgraph.

    """
    denomin = "{" \
        + "".join("%s, "
                  % propa[3]['qp_state']
                  for propa
                  in start_graph.in_edges(subgraph, keys=True, data=True)
                  if not subgraph.has_edge(propa[0], propa[1], propa[2])) \
        + "".join("%s, "
                  % propa[3]['qp_state']
                  for propa
                  in start_graph.out_edges(subgraph, keys=True, data=True)
                  if not subgraph.has_edge(propa[0], propa[1], propa[2]))
    denomin = denomin.strip(', ') + '}'
    return denomin


class MbptDiagram(adg.diag.Diagram):
    """Describes a MBPT diagram with its related properties.

    Attributes:
        incidence (NumPy array): The incidence matrix of the graph.
        excitation_level (int): The single, double, etc., excitation character.
        complex_conjugate (int): The tag number of the diagram's complex
            conjugate. -1 is the graph has none.
        expr (str): The MBPT expression associated to the diagram.
        cd_expr (str): The expression associated to the diagram in a
            computer-readable format.
        adjacency_mat (NumPy array): The adjacency matrix of the graph.

    """

    __slots__ = ('incidence', 'excitation_level', 'complex_conjugate', 'expr',
                 'cd_expr', 'adjacency_mat')

    def __init__(self, mbpt_graph, tag_num):
        """Generate a MBPT diagram using the appropriate NetworkX graph.

        Args:
            mbpt_graph (NetworkX MultiDiGraph): The actual diagram.
            tag_num (int): The tag number associated to the graph.

        """
        adg.diag.Diagram.__init__(self, mbpt_graph)
        self.tags = [tag_num]
        # Beware of the sign convention !!!
        self.incidence = - nx.incidence_matrix(self.graph,
                                               oriented=True).todense()
        self.attribute_ph_labels()
        self.attribute_expression()
        self.excitation_level = self.calc_excitation()
        self.complex_conjugate = -1
        self.loops_number()
        self.adjacency_mat = nx.to_numpy_array(self.graph, dtype=int)

    def attribute_expression(self):
        """Initialize the expression associated to the diagram."""
        sign = "-" if (self.count_hole_lines()
                       - self.loops_number()) % 2 == 1 else ""
        eq_lines = np.array(self.incidence.transpose())
        neq_lines = np.asarray(list(i for i in set(map(tuple, eq_lines))))
        nedges_eq = 2**(len(eq_lines)-len(neq_lines))

        self.expr = sign \
            + ("\\dfrac{1}{%i}" % nedges_eq if nedges_eq != 1 else "") \
            + "\\sum{\\dfrac{%s}{%s}}\n" % (self.extract_numerator(),
                                            self.extract_denominator())
        self.cd_expr = "{%i, {%s}, {%s}};" % (nedges_eq,
                                              self.cd_numerator(),
                                              self.cd_denominator())

    def calc_excitation(self):
        """Return an integer coding for the excitation level of the diag.

        Returns:
            (int): The singles / doubles / etc. character of the graph.

        """
        max_excited_states = 0
        for row in range(1, self.graph.number_of_nodes()):
            nb_excited_states = sum(1 for col
                                    in range(self.graph.number_of_edges())
                                    if self.incidence[0:row, col].sum() == 1)
            if nb_excited_states > max_excited_states \
                    and nb_excited_states != 2:
                max_excited_states = nb_excited_states
        return max_excited_states if max_excited_states != 0 else 2

    def count_hole_lines(self):
        """Return an integer for the number of hole lines in the graph.

        Returns:
            (int): The number of holes in the diagram.

        """
        return sum(1 for edge in self.graph.edges() if edge[0] > edge[1])

    def is_complex_conjug_of(self, test_diagram):
        """Return True if self and test_diagram are complex conjugate.

        Args:
            test_diagram (MbptDiagram): A diagram to compare with.

        Return:
            (bool): The complex conjugate status of the pair of diagrams.

        """
        # Check the adjacency mat against the anti-transposed one of test_diag
        return np.array_equal(self.adjacency_mat,
                              test_diagram.adjacency_mat[::-1, ::-1].T)

    def attribute_ph_labels(self):
        """Attribute the appropriate qp labels to the graph's propagators."""
        labels = list(string.ascii_lowercase)
        # Labelling needs to be shifted for higher orders
        if len(self.graph) < 6:
            h_labels, p_labels = labels[0:8], labels[8:]
        else:
            h_labels, p_labels = labels[0:13], labels[13:]
        for prop in self.graph.edges(keys=True, data=True):
            prop[3]['qp_state'] = h_labels.pop(0) if prop[0] < prop[1] \
                else p_labels.pop(0)

    def extract_denominator(self):
        """Return the denominator for a MBPT graph.

        Returns:
            (str): The denominator of the diagram.

        """
        denominator = ""
        graph = self.graph
        vertices = list(range(1, len(graph)))
        while len(vertices) >= 1:
            denominator += "%s\\ " % adg.diag.extract_denom(
                graph, graph.subgraph(vertices))
            del vertices[0]
        return denominator

    def cd_denominator(self):
        """Return the computer-readable denominator of the graph.

        Return:
            (str): The graph denominator tailored for automated frameworks.

        """
        denominator = ""
        graph = self.graph
        vertices = list(range(1, len(graph)))
        while len(vertices) >= 1:
            denominator += "%s, " % extract_cd_denom(graph,
                                                     graph.subgraph(vertices))
            del vertices[0]
        return denominator.strip(', ')

    def extract_numerator(self):
        """Return the numerator associated to a MBPT graph.

        Returns:
            (str): The numerator of the diagram.

        """
        graph = self.graph
        numerator = ""
        for vertex in graph:
            # First add the qp states corresponding to propagators going out
            numerator += "v_{" + "".join(
                prop[3]['qp_state']
                for prop in graph.out_edges(vertex, keys=True, data=True))
            # Add the qp states corresponding to propagators coming in
            numerator += "".join(
                prop[3]['qp_state']
                for prop in graph.in_edges(vertex, keys=True, data=True)) \
                + "} "
        return numerator

    def cd_numerator(self):
        """Return the computer-readable numerator.

        Returns:
            (str): The graph numerator tailored for automated frameworks.

        """
        graph = self.graph
        numerator = ""
        for vertex in graph:
            # Attribute the correct operator to each vertex
            numerator += "{"
            # First add the qp states corresponding to propagators going out
            numerator += ", ".join(
                prop[3]['qp_state']
                for prop in graph.out_edges(vertex, keys=True, data=True))
            numerator += ', '
            # Add the qp states corresponding to propagators coming in
            numerator += ", ".join(
                prop[3]['qp_state']
                for prop in graph.in_edges(vertex, keys=True, data=True))
            numerator = numerator.strip(', ') + "}, "
        return numerator.strip(', ')

    def loops_number(self):
        """Return the number of loops in the diagram as an integer.

        Returns:
            (int): The number of loops in the graph.

        """
        nb_loops = 0
        nb_checked_props = 0
        diag = self.graph
        nx.set_edge_attributes(diag, False, 'checked')
        while nb_checked_props < diag.number_of_edges():
            prop = list(edge for edge in diag.edges(keys=True, data=True)
                        if edge[3]['checked'] is False)[0]
            while prop[3]['checked'] is False:
                prop[3]['checked'] = True
                nb_checked_props += 1
                left_right_label = list(diag.in_edges(prop[1],
                                                      keys=True,
                                                      data=True)).index(prop)
                prop = list(diag.out_edges(prop[1],
                                           keys=True,
                                           data=True))[left_right_label]
            nb_loops += 1
        return nb_loops

    def write_section(self, result, commands, flags):
        """Write sections for MBPT result file.

        Args:
            result (file): The LaTeX output file to be written in.
            commands (dict): The flags associated with run management.
            flags (dict): The identifier of each section-starting graph.

        """
        if self.tags[0] == flags['singles']:
            result.write("\\section{Singles}\n\n")
        elif self.tags[0] == flags['doubles']:
            result.write("\\section{Doubles}\n\n")
        elif self.tags[0] == flags['triples']:
            result.write("\\section{Triples}\n\n")
        elif self.tags[0] == flags['quadruples']:
            result.write("\\section{Quadruples}\n\n")
        elif self.tags[0] == flags['quintuples+']:
            result.write("\\section{Quintuples and higher}\n\n")
        result.write("\\paragraph{Diagram %i:}\n" % (self.tags[0] + 1))
        if self.complex_conjugate >= 0:
            result.write("Complex conjugate diagram: %i\n"
                         % (self.complex_conjugate + 1))
        write_diag_exp(result, self)



"""Routines handling the run of ADG."""
from __future__ import print_function

from future import standard_library
standard_library.install_aliases()
from builtins import input
from builtins import range
from adg.tools import reversed_enumerate
import os
import argparse
import shutil
import networkx as nx
import adg.mbpt
import adg.bmbpt
import adg.pbmbpt
import adg.diag


def parse_command_line(cli_args):
    """Return run commands from the Command Line Interface.

    Args:
        cli_args: Command-line arguments submitted with the program.

    Returns:
        (Namespace): Appropriate commands to manage the program's run.

    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="AUTOMATIC DIAGRAM GENERATOR v%s\n\n" % adg.__version__
        + "Generates diagrams at a given order for a "
        + "range of many-body formalisms",
        epilog="If a theory is chosen in batch mode, all flags associated\n"
        + "to other theories will be deactivated to avoid conflicts.\n\n")

    basic_args = parser.add_argument_group(
        title="Basic arguments",
        description="For ADG to run, you need either to run it in\n"
        + "interactive mode or to provide a theory of interest and an order\n"
        + "at which to run the calculation.")
    run_args = parser.add_argument_group(
        title="Run arguments",
        description="Those arguments control the content of the ouput files.\n"
        )
    mbpt_args = parser.add_argument_group(
        title="MBPT-specific arguments",
        description="Arguments available only for MBPT calculations.\n")
    bmbpt_args = parser.add_argument_group(
        title="BMBPT-specific arguments",
        description="Arguments available only for (P)BMBPT calculations.\n")

    basic_args.add_argument(
        "-o", "--order", type=int, nargs='+', choices=list(range(1, 10)),
        help="order of the diagrams (>=1) or N_A, N_B, N_C truncation for BIMSRG")
    basic_args.add_argument(
        "-t", "--theory", type=str, choices=['MBPT', 'BMBPT', 'PBMBPT', 'BIMSRG'],
        help="theory of interest: MBPT, BMBPT, PBMBPT, BIMSRG")
    basic_args.add_argument(
        "-i", "--interactive", action="store_true",
        help="execute ADG in interactive mode")

    bmbpt_args.add_argument(
        "-can", "--canonical", action="store_true",
        help="consider only canonical diagrams")
    bmbpt_args.add_argument(
        "-nobs", "--nbody_observable", type=int, choices=list(range(1, 4)), default=2,
        help="maximal n-body character of the observable [1-3], default = 2")
    bmbpt_args.add_argument(
        "-3NF", "--with_3NF", action="store_true",
        help="use two and three-body forces for (P)BMBPT diagrams")
    bmbpt_args.add_argument(
        "-dt", "--draw_tsds", action="store_true",
        help="draw Time-Structure Diagrams (BMBPT or PBMBPT)")

    run_args.add_argument(
        "-d", "--draw_diags", action="store_true",
        help="draw the diagrams using FeynMF")
    run_args.add_argument(
        "-c", "--compile", action="store_true",
        help="compile the LaTeX output file with PDFLaTeX")

    mbpt_args.add_argument(
        "-cd", "--cd_output", action="store_true",
        help="produce computer-readable output for automated framework (MBPT)")

    args = parser.parse_args(cli_args)

    if (not args.interactive) and ((args.order is None)
                                   or (args.theory is None)):
        print("\nPlease either run the interactive mode, or the batch mode by")
        print("providing the theory and the order for the desired diagrams.\n")
        print("Use 'adg -h' for help.\n")
        exit()

    # Avoid conflicting flags
    if args.theory not in ('BMBPT', 'PBMBPT') and not args.interactive:
        args.canonical = None
        args.with_3NF = None
        args.nbody_observable = 2
        args.draw_tsds = None
    if args.theory != 'MBPT' and not args.interactive:
        args.cd_output = None
    if args.theory != 'BIMSRG' and len(args.order) > 1:
        print("Order of the diagrams has to be a single int. Exiting.")
        exit()

    if args.theory == 'BIMSRG':
        if len(args.order) == 1:
            print("Using the traditional IMSRG truncation: N_A = N_B = N_C.")
            dummy_order = args.order[0]
            args.order = (dummy_order, dummy_order, dummy_order)
        elif len(args.order) == 2:
            print("Truncate commutator and operators differently: N_A = N_B.")
            ope_order, comm_order = args.order[0], args.order[-1]
            args.order = (ope_order, ope_order, comm_order)
        elif len(args.order) == 3:
            print("Using the N_A, N_B, N_C truncation scheme.")
            a_order, b_order, comm_order = args.order[0], args.order[1], args.order[-1]
            args.order = (a_order, b_order, comm_order)
        else:
            print("Truncation order has to be three int at most. Exiting.")
            exit()
        if args.draw_diags and max(args.order) >= 7:
            args.draw_diags = False
            print("Diagram drawing is not supported for BIMSRG above order 6.")
            print("The '-d' flag has been deactivated.")
    else:
        dummy_order = args.order[0]
        args.order = dummy_order

    return args


def interactive_interface(commands):
    """Run the interactive interface mode, return the appropriate commands.

    Args:
        commands (Namespace): Flags for the run management.

    Returns:
        (Namespace): Flags initialized through keyboard input.

    """
    theories = ["BMBPT", "MBPT", "PBMBPT", "BIMSRG"]

    commands.theory = input('MBPT, BMBPT, PBMBPT or BIMSRG?\n').upper()
    while commands.theory not in theories:
        print("Invalid theory!")
        commands.theory = input('MBPT, BMBPT, PBMBPT or BIMSRG?\n').upper()

    if commands.theory != 'BIMSRG':
        try:
            commands.order = int(input('Order of the diagrams? [1-9]\n'))
        except ValueError:
            print("Please enter an integer value! Program exiting.")
            exit()
        while commands.order < 1 or commands.order > 9:
            print("Perturbative order too small or too high!")
            commands.order = int(input('Order of the diagrams? [1-9]\n'))
    else:
        commands.order = [0, 0, 0]
        for position, operator in enumerate('A', 'B', 'C'):
            commands.order[position] = get_bimsrg_truncation_order(operator)
        commands.order = tuple(commands.order)

    if commands.theory in ("BMBPT", "PBMBPT"):
        commands.canonical = input(
            "Consider only canonical diagrams? (y/N)").lower() == 'y'
        commands.with_3NF = input(
            "Include three-body forces? (y/N)").lower() == 'y'
        try:
            commands.nbody_observable = int(input(
                "Maximal n-body character of the observable? [1-3]"))
        except ValueError:
            print("Please enter an integer value! Program exiting.")
            exit()
        while commands.nbody_observable < 1 or commands.nbody_observable > 3:
            print("The observable must be 1-body, 2-body or 3-body!")
            commands.nbody_observable = int(input(
                "Maximal n-body character of the observable? [1-3]"))
        commands.draw_tsds = input(
            "Draw time-structure diagrams? (y/N)").lower() == 'y'

    commands.draw_diags = input(
        "Generate diagrams FeynMF instructions in TeX file? (y/N) "
        ).lower() == 'y'

    if commands.theory == "BIMSRG" and commands.draw_diags \
            and max(commands.order) >= 7:
        commands.draw_diags = False
        print("Diagram drawing is not supported for BIMSRG above order 6.")
        print("The '-d' flag has been deactivated.")

    if commands.theory == "MBPT":
        commands.cd_output = input(
            "Produce a CD output file? (y/N) ").lower() == 'y'
    commands.compile = input("Compile pdf? (y/N) ").lower() == 'y'

    return commands


def get_bimsrg_truncation_order(operator):
    """Return the truncation order of a given operator from the user input.

    Args:
        operator (str): The letter corresponding to the operator name.py

    Returns:
        order (int): The truncation rank of the operator.
    """
    try:
        order = int(input('Order of the %s operator? [1-9]\n' % operator))
    except ValueError:
        print("Please enter an integer value! Program exiting.")
        exit()
    while order < 1 or order > 9:
        print("Order too small or too high!")
        order = int(input('Order of the %s operator? [1-9]\n' % operator))
    return order


def attribute_directory(commands):
    """Create missing directories and return the working directory.

    Args:
        commands (Namespace): Flags for the run management.

    Returns:
        (str): Path to the result folder.

    >>> com = argparse.Namespace()
    >>>
    >>> com.theory, com.order = 'BMBPT', 4
    >>> com.with_3NF, com.nbody_observable, com.canonical = False, 2, False
    >>>
    >>> attribute_directory(com)
    'BMBPT/Order-4_2body_observable'
    >>>
    >>> com.theory, com.order = 'BMBPT', 5
    >>> com.with_3NF, com.nbody_observable, com.canonical = True, 3, False
    >>>
    >>> attribute_directory(com)
    'BMBPT/Order-5_3body_observable_with3N'
    >>>
    >>> com.theory, com.order = 'MBPT', 3
    >>> com.with_3NF, com.nbody_observable, com.canonical = False, 2, False
    >>>
    >>> attribute_directory(com)
    'MBPT/Order-3'
    >>>
    >>> com.theory, com.order = 'BIMSRG', (1,2,3)
    >>>
    >>> attribute_directory(com)
    'BIMSRG/Order_1_2_3'

    """
    directory = '%s/' % commands.theory
    if commands.theory != "BIMSRG":
        directory += 'Order-%i' % commands.order
        if commands.canonical:
            directory += '_canonical'
        if commands.theory in ('BMBPT', 'PBMBPT'):
            directory += '_%ibody_observable' % commands.nbody_observable
        if commands.with_3NF:
            directory += '_with3N'
    else:
        directory += 'Order_%i_%i_%i' % commands.order
    if not os.path.exists(directory):
        os.makedirs(directory)
    if not os.path.exists(directory+"/Diagrams"):
        os.makedirs(directory+"/Diagrams")
    return directory


def generate_diagrams(commands, id_generator):
    """Return a list with diagrams of the appropriate type.

    Args:
        commands (Namespace): Flags for the run management.
        id_generator (UniqueID): A unique ID number generator.

    Returns:
        (list): All the diagrams of the appropriate Class and order.

    """
    if commands.theory == "MBPT":
        diagrams = adg.mbpt.diagrams_generation(commands.order)
    elif commands.theory in ("BMBPT", "PBMBPT"):
        diagrams = adg.bmbpt.diagrams_generation(commands.order,
                                                 commands.with_3NF,
                                                 commands.nbody_observable,
                                                 commands.canonical)
    elif commands.theory == "BIMSRG":
        diagrams, switch_flag = adg.bimsrg.diagrams_generation(commands.order)
    else:
        print("Invalid theory! Exiting program.")
        exit()
    print("Number of matrices produced: ", len(diagrams))

    diags = [nx.from_numpy_array(diagram, create_using=nx.MultiDiGraph(),
                                 parallel_edges=True) for diagram in diagrams]

    if commands.theory == "MBPT":
        for i_diag, diag in reversed_enumerate(diags):
            if (nx.number_weakly_connected_components(diag)) != 1:
                del diags[i_diag]

    adg.diag.label_vertices(diags,
                            commands.theory,
                            switch_flag if commands.theory == 'BIMSRG' else -1)

    if commands.theory in ('BMBPT', "PBMBPT"):
        diagrams = [adg.bmbpt.BmbptFeynmanDiagram(graph, id_generator.get())
                    for graph in diags]
    elif commands.theory == 'MBPT':
        diagrams = [adg.mbpt.MbptDiagram(graph, id_generator.get())
                    for graph in diags]
    elif commands.theory == "BIMSRG":
        diagrams = [adg.bimsrg.BimsrgDiagram(graph, id_generator.get())
                    for graph in diags]

    if commands.theory == "PBMBPT":
        for idx, diagram in reversed_enumerate(diagrams):
            new_graphs = adg.pbmbpt.generate_anomalous_diags(
                diagram,
                3 if commands.with_3NF else 2
            )
            new_diags = [adg.pbmbpt.ProjectedBmbptDiagram(diag,
                                                          id_generator.get(),
                                                          idx,
                                                          spawn_idx)
                         for spawn_idx, diag in enumerate(new_graphs)]
            del diagrams[idx]
            adg.pbmbpt.filter_new_diagrams(new_diags, diagrams)
            diagrams += new_diags
    return diagrams


def order_diagrams(diagrams, commands):
    """Return the ordered unique diagrams with a dict of numbers per type.

    Args:
        diagrams (list): The diagrams of the appropriate Class.
        commands (Namespace): Flags for the run management.

    Returns:
        (tuple): First element is the list of ordered and unique diagrams.
        Second element is a dict with the number of diagrams per type. Third
        element is a dict with the identifiers of diagrams starting each output
        file section.

    """
    if commands.theory in ("BMBPT", "PBMBPT"):
        diagrams, diag_nbs, section_flags = adg.bmbpt.order_diagrams(diagrams)
    elif commands.theory == "MBPT":
        diagrams, diag_nbs, section_flags = adg.mbpt.order_diagrams(diagrams)
    elif commands.theory == "BIMSRG":
        order = max(commands.order)
        diagrams, diag_nbs, section_flags = adg.bimsrg.order_diagrams(diagrams,
                                                                      order)

    # Reattribute a number to the BMBPT diagrams
    if commands.theory == "BMBPT":
        for ind, diagram in enumerate(diagrams):
            diagram.tags[0] = ind

    return diagrams, diag_nbs, section_flags


def print_diags_numbers(commands, diags_nbs):
    """Print the number of diagrams for each major type.

    Args:
        commands (Namespace): Flags for the run management.
        diags_nbs (dict): The number of diagrams for each major type.

    """
    print("Number of connected diagrams: ", diags_nbs['nb_diags'])

    if commands.theory in ("BMBPT", "PBMBPT"):
        print(
            "\n2N valid diagrams: %i\n" % diags_nbs['nb_2']
            + "2N energy canonical diagrams: %i\n" % diags_nbs['nb_2_hf']
            + "2N canonical diagrams for a generic operator only: %i"
            % diags_nbs['nb_2_ehf']
        )
        if not commands.canonical:
            print(
                "2N non-canonical diagrams: %i" % diags_nbs['nb_2_not_hf']
            )
        if commands.with_3NF:
            print(
                "\n3N valid diagrams: %i\n" % diags_nbs['nb_3']
                + "3N energy canonical diagrams: %i\n" % diags_nbs['nb_3_hf']
                + "3N canonical diagrams for a generic operator only: %i"
                % diags_nbs['nb_3_ehf']
            )
            if not commands.canonical:
                print(
                    "3N non-canonical diagrams: %i" % diags_nbs['nb_3_not_hf']
                )
    elif commands.theory == "MBPT":
        print(
            "\nValid diagrams: %i\n\n" % diags_nbs['nb_diags']
            + "Singles: %i\n" % diags_nbs['singles']
            + "Doubles: %i\n" % diags_nbs['doubles']
            + "Triples: %i\n" % diags_nbs['triples']
            + "Quadruples: %i\n" % diags_nbs['quadruples']
            + "Quintuples and higher: %i" % diags_nbs['quintuples+']
        )
    elif commands.theory == "BIMSRG":
        print("\nValid diagrams: %i" % diags_nbs['nb_diags'])
        for n in range(1, commands.order[-1] + 1):
            print("d_max = %i diagrams: %i" % (n, diags_nbs[n]))
    print()


def prepare_drawing_instructions(directory, commands, diagrams, diagrams_time):
    """Write FeynMP files for the different diagrams.

    Args:
        directory (str): Path to the output folder.
        commands (Namespace): Flags for the run management.
        diagrams (list): All the diagrams of interest.
        diagrams_time (list): All the associated TSDs if appropriate.

    """
    create_feynmanmp_files(diagrams, commands.theory, directory, 'diag')
    if commands.draw_tsds:
        create_feynmanmp_files(diagrams_time, commands.theory,
                               directory, 'time')


def create_feynmanmp_files(diagrams, theory, directory, diag_type):
    """Create and move the appropriate feynmanmp files to the right place.

    Args:
        diagrams (list): The studied diagrams.
        theory (str): Name of the theory of interest.
        directory (str): Path to the result folder.
        diag_type (str): Type of studied diagrams used for drawing.

    """
    for diag in diagrams:
        if theory == "PBMBPT" and diag_type == 'diag':
            diag_name = '%s_%i_%i' % (diag_type, diag.tags[0], diag.tags[1])
        else:
            diag_name = '%s_%i' % (diag_type, diag.tags[0])
        adg.diag.feynmf_generator(diag.graph,
                                  'MBPT' if diag_type == 'time' else theory,
                                  diag_name)
        shutil.move('%s.tex' % diag_name,
                    "%s/Diagrams/%s.tex" % (directory, diag_name))


def write_file_header(latex_file, commands, diags_nbs):
    """Write the header of the result tex file.

    Args:
        latex_file (file): LaTeX output file of the program.
        commands (Namespace): Flags to manage the program's run.
        diags_nbs (dict): Number of diagrams per major type.

    """
    header = "\\documentclass[10pt,a4paper]{article}\n" \
        + "\\usepackage[left=2cm,right=2cm," \
        + "bottom=2.5cm,top=2.5cm]{geometry}\n" \
        + "\\usepackage[utf8]{inputenc}\n" \
        + "\\usepackage[T1]{fontenc}\n" \
        + "\\usepackage{lmodern}\n" \
        + "\\usepackage[hyperindex=true]{hyperref}" \
        + "\\usepackage[english]{babel}\n\\usepackage{amsmath}\n" \
        + "\\usepackage{amsfonts}\n\\usepackage{amssymb}\n"
    if commands.draw_diags:
        header = "%s\\usepackage{feynmp-auto}\n" % header
    if commands.theory in ('BMBPT', 'PBMBPT') and commands.order >= 3:
        header = "%s\\usepackage[landscape]{geometry}\n" % header

    if commands.theory != 'BIMSRG':
        header = header \
            + "\\title{Diagrams and algebraic expressions at order %i in %s}\n" \
            % (commands.order, commands.theory) \
            + "\\author{The ADG Dev Team}\n"
    else:
        header = header \
            + "\\allowdisplaybreaks\n\n" \
            + "\\title{Diagrams and algebraic expressions at order (%i,%i;%i) in BIMSRG}\n" \
            % commands.order \
            + "\\author{The ADG Dev Team}\n"
    latex_file.write("%s\n\\begin{document}\n\n\\maketitle\n\n" % header)

    if commands.theory in ("BMBPT", "PBMBPT"):
        adg.bmbpt.write_header(latex_file, commands, diags_nbs)
    elif commands.theory == "MBPT":
        adg.mbpt.write_header(latex_file, diags_nbs)
    elif commands.theory == "BIMSRG":
        adg.bimsrg.write_header(latex_file, commands, diags_nbs)

    latex_file.write("\n\\tableofcontents\n\n")

    if commands.theory == "BIMSRG":
        adg.bimsrg.write_permutator_section(latex_file, commands)


def compile_manager(directory):
    """Compile the program's LaTeX ouput file.

    Args:
        directory (str): Path to the ouput folder.

    """
    os.chdir(directory)
    os.system("pdflatex -shell-escape -interaction=batchmode result.tex")
    # Second compilation for table of contents and diagrams
    os.system("pdflatex -shell-escape -interaction=batchmode result.tex")
    os.chdir("../..")


def clean_folders(directory, commands):
    """Delete temporary files and folders.

    Args:
        directory (str): Path to the output folder.
        commands (Namespace): Flags to manage the program's run.

    """
    os.chdir(directory)
    if commands.draw_diags:
        # Remove the temporary Diagrams folder
        shutil.rmtree('Diagrams')
        # Get rid of undesired feynmp files to keep a clean directory
        for filename in os.listdir('.'):
            if filename.startswith("time") or filename.startswith("diag") \
                    or filename.startswith("equivalent"):
                os.unlink(filename)

    if commands.compile:
        # Remove LaTeX auxiliary files
        os.unlink("result.aux")
        os.unlink("result.log")
        os.unlink("result.out")
        os.unlink("result.toc")
    os.chdir("../..")



"""Module with functions relative to time-stucture diagrams, called by ADG."""

from builtins import range
from adg.tools import reversed_enumerate
import copy
import os
import math
import networkx as nx
import adg.diag


def time_structure_graph(diag):
    """Return the time-structure graph associated to the graph.

    Args:
        diag (BmbptFeymmanDiagram): The BMBPT graph of interest.

    Returns:
        (NetworkX MultiDiGraph): The time-structure diagram.

    """
    import adg.pbmbpt
    time_graph = diag.graph.to_directed()
    if time_graph.nodes[0]['operator']:
        time_graph.add_edges_from((0, vertex)
                                  for vertex in range(1, len(time_graph)))
    if isinstance(diag, adg.pbmbpt.ProjectedBmbptDiagram):
        edges_copy = copy.deepcopy(time_graph.edges(keys=True, data=True))
        time_graph.remove_edges_from((edge[0], edge[1], edge[2])
                                     for edge in edges_copy
                                     if 'anomalous' in edge[3]
                                     and edge[3]['anomalous'])
    return adg.diag.to_skeleton(time_graph)


def tree_time_structure_den(time_graph):
    """Return the denominator associated to a tree time-structure graph.

    Args:
        time_graph (NetworkX MultiDiGraph): The TSD of interest.

    Returns:
        (str): The denominator associated to the TSD.

    """
    denominator = ""
    i = 0
    for vertex in time_graph:
        if not time_graph.nodes[vertex]['operator']:
            time_graph.nodes[vertex]['label'] = 'a_%i' % (i + 1)
            i += 1
    for vertex in time_graph:
        if not time_graph.nodes[vertex]['operator']:
            if time_graph.out_degree(vertex) == 0:
                denominator += time_graph.nodes[vertex]['label']
            else:
                denominator += "(%s" % time_graph.nodes[vertex]['label'] \
                    + "".join("+ %s" % time_graph.nodes[descendant]['label']
                              for descendant
                              in nx.descendants(time_graph, vertex)) + ")"
    return denominator


def equivalent_labelled_tsds(equivalent_trees, labelled_tsds):
    """Return the list of labelled TSDs corresponding to equivalent TSDs.

    Args:
        equivalent_trees (list): The equivalent tree TSDs of a non-tree TSD.
        labelled_tsds (list): The labelled TSDs obtained from BMBPT diagrams.

    Returns:
        (str): The list of tag numbers of the equivalent TSDs.

    """
    op_nm = nx.algorithms.isomorphism.categorical_node_match('operator', False)
    eq_labelled_tsds = ""
    for eq_tree_graph in equivalent_trees:
        for comp_tdiag in labelled_tsds:
            if tuple(sorted(tuple((eq_tree_graph.in_degree(node),
                                   eq_tree_graph.out_degree(node))
                                  for node in eq_tree_graph))) \
                    == comp_tdiag.io_degrees \
                    and comp_tdiag.is_tree:
                if nx.is_isomorphic(eq_tree_graph, comp_tdiag.graph, op_nm):
                    eq_labelled_tsds += " T%s," % (comp_tdiag.tags[0]+1)
                    break
    return "".join("%s." % eq_labelled_tsds.strip(','))


def write_section(latex_file, directory, pdiag, time_diagrams, nb_tree_tsds,
                  diagrams):
    """Write the appropriate section for tsd diagrams in the LaTeX file.

    Args:
        latex_file (file): The LaTeX output file of the program.
        directory (str): Path to the output folder.
        pdiag (bool): ``True`` if diagrams are to be drawn.
        time_diagrams (list): The ensemble of TSDs.
        nb_tree_tsds (int): Number of tree TSDs.
        diagrams (list): All produced BmbptFeymmanDiagrams.

    """
    latex_file.write("\\section{Time-structure diagrams}\n\n"
                     + "\\subsection{Tree diagrams}\n\n")
    for tdiag in time_diagrams:
        if tdiag.tags[0] == nb_tree_tsds:
            latex_file.write("\\subsection{Non-tree diagrams}\n\n")
        latex_file.write("\\paragraph{Time-structure diagram T%i:}\n"
                         % (tdiag.tags[0]+1))
        if pdiag:
            with open(directory
                      + "/Diagrams/time_%i.tex" % tdiag.tags[0]) as time_file:
                latex_file.write('\n\\begin{center}\n%s\n\\end{center}\n\n'
                                 % time_file.read())
        latex_file.write("\\begin{equation}\n\\text{T%i} = "
                         "%s\\end{equation}\n\n"
                         % (tdiag.tags[0]+1, tdiag.expr))
        if tdiag.is_tree:
            latex_file.write("Resummation power: %i\n\n" % tdiag.resum)
        else:
            latex_file.write("Equivalent tree diagrams: %s\n\n"
                             % equivalent_labelled_tsds(tdiag.equivalent_trees,
                                                        time_diagrams))
            latex_file.write('\n\\begin{center}\n')
            tdiag.draw_equivalent_tree_tsds(latex_file)
            latex_file.write('\n\\end{center}\n\n')
        latex_file.write("Number of related Feynman diagrams: %i.\n\n"
                         % (len(tdiag.tags)-1))
        latex_file.write("Related Feynman diagrams:%s.\n\n"
                         % tdiag.get_feynman_diags(diagrams))


def disentangle_cycle(time_graph, cycle_nodes):
    """Separate a cycle in a sum of tree diagrams.

    Args:
        time_graph (NetworkXn MultiDiGraph): A time-structure diagram.
        cycle_nodes (tuple): Integers encoding the positions of the end nodes
            of the cycle.

    Returns:
        (list): New graphs produced from treating the cycles in the TSD.

    """
    paths = list(nx.all_simple_paths(time_graph,
                                     cycle_nodes[0],
                                     cycle_nodes[1]))
    new_graphs = []
    insert_node = paths[0][1]
    for daughter_node in paths[1][1:]:
        new_graph = time_graph.to_directed()
        new_graph.add_edge(insert_node, daughter_node)
        for test_node in paths[1]:
            if test_node in list(time_graph.predecessors(daughter_node)):
                mother_node = test_node
                break
        new_graph.add_edge(mother_node, insert_node)
        adg.diag.to_skeleton(new_graph)
        new_graphs.append(new_graph)
    return new_graphs


def find_cycle(graph):
    """Return start and end nodes for an elementary cycle.

    Args:
        graph (NetworkX MultiDiGraph): A TSD with cycle(s) to be treated.

    Returns:
        (tuple): Positions of the two end nodes of a cycle in the graph.

    """
    cycle_found = False
    for node_a in (node for node in graph if graph.out_degree(node) >= 2):
        for node_b in (node for node in graph if graph.in_degree(node) >= 2):
            paths = list(nx.all_simple_paths(graph, node_a, node_b))
            if len(paths) >= 2:
                cycle_nodes = (node_a, node_b)
                cycle_found = True
                # Avoid false positive when node_a has 2+ branches out
                # but only one goes to node_b
                for test_node in paths[0][1:-1]:
                    if test_node in paths[1][1:-1]:
                        cycle_found = False
                        break
                break
        if cycle_found:
            break
    return cycle_nodes


def treat_tsds(diagrams_time):
    """Order TSDs, produce their expressions, return also number of trees.

    Args:
        diagrams_time (list): All the associated TSDs.

    Returns:
        (tuple): List of TSDs, number of tree TSDs

    """
    tree_tsds = []
    for i_diag, diag in reversed_enumerate(diagrams_time):
        if diag.is_tree:
            tree_tsds.append(diag)
            del diagrams_time[i_diag]

    adg.diag.topologically_distinct_diagrams(tree_tsds)
    adg.diag.topologically_distinct_diagrams(diagrams_time)

    diagrams_time = tree_tsds + diagrams_time

    for index, t_diag in enumerate(diagrams_time):
        t_diag.tags.insert(0, index)
        if not t_diag.is_tree:
            t_diag.equivalent_trees = t_diag.treat_cycles()
            t_diag.expr = " + ".join("\\frac{1}{%s}"
                                     % adg.tsd.tree_time_structure_den(graph)
                                     for graph
                                     in t_diag.equivalent_trees)
    return diagrams_time, len(tree_tsds)


class TimeStructureDiagram(adg.diag.Diagram):
    """Describes a time-structure diagram with its related properties.

    Attributes:
        perms (dict): The permutations on the vertices for all the BMBPT
            diagrams associated to this TSD.
        equivalent_trees (list): The tag numbers of the equivalent tree TSDs
            associated to a non-tree TSD.
        is_tree (bool): The tree or non-tree character of a TSD.
        expr (str): The Goldstone denominator associated to the TSD.
        resum (int): The resummation power of a tree TSD.

    """

    __slots__ = ('perms', 'equivalent_trees', 'is_tree', 'expr', 'resum')

    def __init__(self, bmbpt_diag):
        """Generate a tsd diagram out of a BMBPT one.

        Args:
            bmbpt_diag (BmbptFeynmanDiagram): The BMBPT graph used to be
                turned into a TSD.

        """
        adg.diag.Diagram.__init__(self, time_structure_graph(bmbpt_diag))
        self.tags = [bmbpt_diag.unique_id]
        self.perms = {bmbpt_diag.unique_id: {i: i
                                             for i in range(len(self.graph))}}
        self.equivalent_trees = []
        if nx.is_arborescence(self.graph):
            self.is_tree = True
            self.expr = "\\frac{1}{%s}" % tree_time_structure_den(self.graph)
            self.resum = self.resummation_power()
        else:
            self.is_tree = False
            self.expr = ""
            self.resum = 0

    def treat_cycles(self):
        """Find and treat cycles in a TSD diagram.

        Returns:
            (list): The unique tree TSDs associated to a non-tree TSD.

        """
        graphs = [self.graph]
        tree_graphs = []
        cycles_left = True
        while cycles_left:
            for gr_index, graph in reversed_enumerate(graphs):
                graphs += disentangle_cycle(graph, find_cycle(graph))
                del graphs[gr_index]
            cycles_left = False
            for graph_indx, graph in reversed_enumerate(graphs):
                if nx.is_arborescence(graph):
                    tree_graphs.append(graph)
                    del graphs[graph_indx]
                else:
                    cycles_left = True
        tree_graphs_uniq = []
        for t_graph in tree_graphs:
            for t_graph_uniq in tree_graphs_uniq:
                if nx.edges(t_graph) == nx.edges(t_graph_uniq):
                    break
            # If the TSD is a new one
            else:
                tree_graphs_uniq.append(t_graph)
        return tree_graphs_uniq

    def draw_equivalent_tree_tsds(self, latex_file):
        """Draw the equivalent tree TSDs for a given non-tree TSD.

        Args:
            latex_file (file): The output LaTeX file of the program.

        """
        for index, graph in enumerate(self.equivalent_trees):
            adg.diag.feynmf_generator(graph,
                                      'MBPT',
                                      'equivalent%i_%i' % (self.tags[0],
                                                           index))
            with open("equivalent%i_%i.tex"
                      % (self.tags[0], index)) as diag_file:
                latex_file.write(diag_file.read())
            os.unlink("./equivalent%i_%i.tex" % (self.tags[0], index))

    def resummation_power(self):
        """Calculate the resummation power of the tree TSD.

        Returns:
            (int): The resummation power associated to the TSD.abs

        """
        power = math.factorial(len(self.graph) - 1)

        for node in range(1, len(self.graph)):
            power /= 1 + len(nx.descendants(self.graph, node))

        return power

    def get_feynman_diags(self, feyn_diagrams):
        """Return the list of Feynman diagrams associated to the TSD.

        Args:
            feyn_diagrams (list): All produced BmbptFeymmanDiagrams.

        Returns:
            (str): All the identifiers of associated BmbptFeymmanDiagrams.

        """
        if isinstance(feyn_diagrams[0], adg.pbmbpt.ProjectedBmbptDiagram):
            identifiers = ""
            for tag in self.tags[1:]:
                for diag in feyn_diagrams:
                    if diag.unique_id == tag:
                        identifiers += " %i.%i," % (diag.tags[0]+1,
                                                    diag.tags[1]+1)
                        break
            identifiers = identifiers.strip(',')
        else:
            identifiers = ",".join(" %i" % (tag+1) for tag in self.tags[1:])

        return identifiers



# -*- coding: utf-8 -*-
"""Installation script for the Automated Diagaram Generator.

You can install the program either by running
    pip install <folder>
or
    python setup.py install
"""
from __future__ import print_function


import sys
from setuptools import setup
import adg

main_dependencies = [
    "setuptools"
]

for dep in main_dependencies:
    try:
        __import__(dep)
    except ImportError:
        print(
            "Error: You do not have %s installed, please\n"
            "       install it. For example doing\n"
            "\n"
            "       pip install %s\n" % (dep, dep)
        )
        sys.exit(1)


setup(
    name='adg',
    version=adg.__version__,
    maintainer='Pierre Arthuis',
    maintainer_email='pierre.arthuis@protonmail.com',
    author=adg.__author__,
    author_email=adg.__email__,
    license=adg.__license__,
    url='https://github.com/adgproject/adg',
    install_requires=[
        "future",
        "more-itertools",
        "networkx>=2.0",
        "numpy",
        "scipy",
    ],
    python_requires='>=2.7.1',
    extras_require=dict(
        # List additional groups of dependencies here (e.g. development
        # dependencies). You can install these using the following syntax:
        # $ pip install -e .[develop]
        develop=[
            'pytest>=3.6',
            'pytest-cov',
            'roman',
            'sphinx',
            'sphinx_rtd_theme',
        ]
    ),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    description='A powerful diagram generator and evaluator for many-body '
                'formalisms in physics and chemistry',
    long_description='ADG is a tool generating diagrams and producing their '
                     'expressions for given many-body formalisms. Diagrammatic '
                     'rules from the formalism are combined with graph theory '
                     'objects to produce diagrams and expressions in a fast, '
                     'simple and error-safe way.\n\n'
                     'The only input consists in the theory and order of '
                     'interest, and the N-body character of the operators of '
                     'interest. The main output is a LaTeX file containing the '
                     'diagrams, their associated expressions and additional '
                     'informations that can be compiled by ADG if needed. '
                     'Other computer-readable files may be produced as well.',
    keywords=[
        'physics',
        'chemistry',
        'theory',
        'many-body',
        'diagram',
        'cli',
        'batch'
    ],
    packages=[
        "adg",
    ],
    data_files=[

        ("share/doc/adg/", [
            "README.md",
        ]),

        ("share/man/man1", [
            "doc/build/man/adg.1",
        ]),

    ],
    entry_points=dict(
        console_scripts=[
            'adg=adg.main:main'
        ]
    ),
    platforms=['linux', 'osx'],
)



"""Routines and class for Projected Bogoliubov MBPT diagrams."""
from __future__ import division

from builtins import str
from builtins import range
from past.utils import old_div
from operator import xor
import copy
import itertools
import networkx as nx
import adg.bmbpt
import adg.tsd
import adg.diag


def generate_anomalous_diags(diag, nbody_max):
    """Generate PBMBPT graphs with anomalous lines, with some redundancy.

    Args:
        diag (BmbptFeynmanDiagram): The diagram to generate children from.
        nbody_max (int): The maximal n-body character of a graph vertex.

    Returns:
        (list): The anomalous graphs generated.

    """
    graph = diag.graph
    anom_graphs = [graph]
    vertices = [vert for vert in graph if not graph.nodes[vert]['operator']]

    nx.set_edge_attributes(graph, False, 'anomalous')

    # Turn normal propagators in anomalous ones
    tweakable_edges = []
    for vert1 in vertices:
        for vert2 in vertices[vert1:]:
            tweakable_edges += [(vert1, vert2) for _
                                in range(graph.number_of_edges(vert1, vert2))]

    for comb in unique_edge_combinations(tweakable_edges,
                                         diag.equivalent_permutations()):
        new_graph = copy.deepcopy(graph)
        for edge in comb:
            key = sum(1 for prop
                      in new_graph.out_edges(edge[0], keys=True, data=True)
                      if edge[1] == prop[1] and not prop[3]['anomalous']) - 1
            new_graph[edge[0]][edge[1]][key]['anomalous'] = True
        anom_graphs.append(new_graph)

    # Loop to generate self-contractions
    for iter_graph in reversed(anom_graphs):
        test_vertices = []
        for v in vertices:
            test_vertices += [v for _
                              in range(nbody_max
                                       - (old_div(iter_graph.degree(v), 2)))]
        if test_vertices:
            for comb in unique_vertex_combinations(
                    test_vertices, equiv_generating_permutations(iter_graph)):
                new_graph = copy.deepcopy(iter_graph)
                new_graph.add_edges_from(((vert, vert) for vert in comb),
                                         anomalous=True)
                anom_graphs.append(new_graph)
    return anom_graphs


def generate_combinations(iter_list):
    """Generate all possible combinations of length 1 to total.

    Attributes:
        iter_list (list): A list of iterable objects.

    Returns:
        (list): A list with all the possible combinations of all lengths.

    >>> print(generate_combinations([1, 2, 3]))
    [(1,), (1, 2), (1, 2, 3), (1, 3), (2,), (2, 3), (3,)]

    """
    combinations = []
    for i in range(1, len(iter_list) + 1):
        combinations += [k for k in itertools.combinations(iter_list, i)]
    # Remove duplicates
    return sorted(set(combinations))


def unique_edge_combinations(edges, permutations):
    """Return all edge combinations not producing equivalent anomalous graphs.

    Attributes:
        edges (list): The edges that can be modified.
        permutations (list): The permutation generating equivalent diagrams.

    Returns:
        (list): The list of edges producing unique anomalous diagrams.

    >>> edges = [(1, 3), (2, 3)]
    >>> permutations = [{1: 1, 2: 2}, {1: 2, 2: 1}]
    >>> print(unique_edge_combinations(edges, permutations))
    [((1, 3), (2, 3)), ((2, 3),)]

    """
    edge_combs = generate_combinations(edges)

    if len(permutations) <= 1:
        return edge_combs

    unique_edge_combs = copy.deepcopy(edge_combs)
    for idx, comb1 in enumerate(edge_combs):
        for comb2 in edge_combs[idx+1:]:
            is_same_perm = False
            if len(comb1) == len(comb2):
                for permutation in permutations[1:]:
                    new_comb = []
                    for edge in comb2:
                        new_0 = permutation[edge[0]] \
                            if edge[0] in permutation.keys() else edge[0]
                        new_1 = permutation[edge[1]] \
                            if edge[1] in permutation.keys() else edge[1]
                        new_comb.append(tuple((new_0, new_1)))
                    if sorted(comb1) == sorted(new_comb):
                        is_same_perm = True
                        break
            if is_same_perm:
                unique_edge_combs.remove(comb1)
                break
    return unique_edge_combs


def unique_vertex_combinations(vertices, permutations):
    """Return vertex combinations generating unique anomalous diagrams.

    Return combinations of vertices on which self-contractions can be added
    without producing topologically equivalent PBMBPT diagrams.

    Attributes:
        vertices (list): Vertices that can be self-contracted.
        permutations (list): The permutations that generate equivalent diags.

    Returns:
        (list): Vertex combinations that do not produce equivalent diags.

    >>> vertices = [1, 3]
    >>> permutations = [{1: 1, 3: 3}, {1: 3, 3: 1}]
    >>> print(unique_vertex_combinations(vertices, permutations))
    [(1, 3), (3,)]

    """
    combinations = generate_combinations(vertices)
    if len(combinations) <= 1:
        return combinations
    unique_combs = copy.deepcopy(combinations)
    for idx, comb1 in enumerate(combinations):
        for comb2 in combinations[idx+1:]:
            is_same_perm = False
            if len(comb1) == len(comb2):
                # Skip the identity
                for permutation in permutations[1:]:
                    new_comb2 = [permutation[item]
                                 if item in permutation.keys() else item
                                 for item in comb2]
                    if sorted(comb1) == sorted(new_comb2):
                        is_same_perm = True
                        break
            if is_same_perm:
                unique_combs.remove(comb1)
                break
    return unique_combs


def equiv_generating_permutations(graph):
    """Return the list of permutations generating equivalent PBMBPT diags.

    Attributes:
        graph (Networkx MultiDiGraph): The graph to be checked.

    Returns:
        (list): The mappings giving equivalent graphs, inc. identity.

    """
    op_nm = nx.algorithms.isomorphism.categorical_node_match('operator', False)
    anom_em = nx.algorithms.isomorphism.categorical_multiedge_match('anomalous',
                                                                    False)
    unsort_io_degrees = []
    for node in graph:
        # Edges going out that are anomalous are annihilators going in
        nb_anom_out_edges = sum(1 for edge
                                in graph.out_edges(node, data=True, keys=True)
                                if edge[3]['anomalous'])
        # And thus should be decounted to NetworkX's out_degree...
        out_degree = graph.out_degree(node) - nb_anom_out_edges
        # ...and added to NetworkX's in_degree
        in_degree = graph.in_degree(node) + nb_anom_out_edges

        unsort_io_degrees.append((in_degree, out_degree))
    unsort_io_degrees = tuple(unsort_io_degrees)

    perm_vertices = [vertex for vertex, degrees in enumerate(unsort_io_degrees)
                     if not graph.nodes[vertex]['operator']
                     and unsort_io_degrees.count(degrees) >= 2]
    permutations = []
    doubled_graph = adg.diag.create_checkable_diagram(graph)
    for permutation in itertools.permutations(perm_vertices):
        permuted_graph = nx.relabel_nodes(doubled_graph,
                                          dict(list(zip(perm_vertices,
                                                        permutation))),
                                          copy=True)
        intersection = copy.deepcopy(doubled_graph)
        intersection.remove_edges_from(e for e in doubled_graph.edges()
                                       if e not in permuted_graph.edges())
        matcher = nx.algorithms.isomorphism.DiGraphMatcher(doubled_graph,
                                                           intersection,
                                                           node_match=op_nm,
                                                           edge_match=anom_em)
        if matcher.is_isomorphic():
            permutations.append(dict(list(zip(perm_vertices, permutation))))
    return permutations


def filter_new_diagrams(new_diags, old_diags):
    """Eliminate diagrams having a topologically equivalent diag.

    Attibutes:
        new_diags (list): The list of newly created PBMBPT diagrams.
        old_diags (list): The list of already checked PBMBPT diagrams.

    """
    iso = nx.algorithms.isomorphism
    op_nm = iso.categorical_node_match('operator', False)
    anom_em = iso.categorical_multiedge_match('anomalous', False)
    for ind, new_diag in adg.tools.reversed_enumerate(new_diags):
        # Only anomalous props that are not self-contractions can produce
        # topologically equivalent diagrams through our generation process
        if new_diag.has_anom_non_selfcontracted_props():
            for new_diag_2 in new_diags[:ind]:
                if new_diag_2.io_degrees == new_diag.io_degrees \
                        and new_diag_2.has_anom_non_selfcontracted_props():
                    matcher = iso.DiGraphMatcher(new_diag.check_graph,
                                                 new_diag_2.check_graph,
                                                 node_match=op_nm,
                                                 edge_match=anom_em)
                    if matcher.is_isomorphic():
                        del new_diags[ind]
                        break
            else:
                # Go through the list backwards: first items are BMBPT diags
                for old_diag in reversed(old_diags):
                    if not isinstance(old_diag, ProjectedBmbptDiagram):
                        # All PBMBPT diags have been parsed
                        break
                    if old_diag.io_degrees == new_diag.io_degrees \
                            and old_diag.has_anom_non_selfcontracted_props():
                        matcher = iso.DiGraphMatcher(old_diag.check_graph,
                                                     new_diag.check_graph,
                                                     node_match=op_nm,
                                                     edge_match=anom_em)
                        if matcher.is_isomorphic():
                            del new_diags[ind]
                            break


class ProjectedBmbptDiagram(adg.bmbpt.BmbptFeynmanDiagram):
    """Describes a PBMBPT diagram with its related properties.

    Attributes:
        two_or_three_body (int): The 2 or 3-body characted of the vertices.
        time_tag (int): The tag number associated to the diagram's
            associated TSD.
        tsd_is_tree (bool): The tree or non-tree character of the
            associated TSD.
        feynman_exp (str): The Feynman expression associated to the diagram.
        diag_exp (str): The Goldstone expression associated to the diagram.
        vert_exp (list): The expression associated to the vertices.
        hf_type (str): The Hartree-Fock, non-Hartree-Fock or Hartree-Fock for
            the energy operator only character of the graph.
        unique_id (int): A unique number associated to the diagram.
        vertex_exchange_sym_factor (int): Lazy-initialized symmetry factor
            associated to the vertex exchange, stored to avoid being computed
            several times.
        check_graph (NetworkX MultiDiGraph): A copy of the graph that can be
            used for topological equivalence checks (lazy-initialized).

    """

    __slots__ = ('_check_graph',)

    def __init__(self, graph, unique_id, tag, child_tag):
        """Generate a PBMBPT diagram by copying a BMBPT one.

        Args:
            graph (NetworkX MultiDiGraph)): The graph of interest.

        """
        adg.bmbpt.BmbptFeynmanDiagram.__init__(self, graph, unique_id)
        self.tags = [tag, child_tag]
        self.set_io_degrees()

    def set_io_degrees(self):
        """Attribute the correct in- and out-degrees to a PBMBPT diagram."""
        unsort_io_degrees = []
        for node in self.graph:
            # Edges going out that are anomalous are annihilators going in
            nb_anom_out_edges = sum(1 for edge
                                    in self.graph.out_edges(nbunch=node,
                                                            data=True,
                                                            keys=True)
                                    if edge[3]['anomalous'])
            # And thus should be decounted to NetworkX's out_degree...
            out_degree = self.graph.out_degree(node) - nb_anom_out_edges
            # ...and added to NetworkX's in_degree
            in_degree = self.graph.in_degree(node) + nb_anom_out_edges

            unsort_io_degrees.append((in_degree, out_degree))
        self.unsort_io_degrees = tuple(unsort_io_degrees)
        self.io_degrees = tuple(sorted(self.unsort_io_degrees))
        self._check_graph = None

    def extract_integral(self):
        """Return the integral part of the Feynman expression of the diag.

        Returns:
            (str): The integral part of its Feynman expression.

        """
        pert_vertex_indices = list(range(1, len(self.graph)))
        integral = "".join("\\mathrm{d}\\tau_%i" % vertex
                           for vertex in pert_vertex_indices)

        normal_props = [prop for prop in self.graph.edges(data='anomalous')
                        if not prop[2]]

        if len(pert_vertex_indices) > 1:
            for vertex_i in pert_vertex_indices:
                integral += "".join("\\theta(\\tau_%i-\\tau_%i) " % (vertex_j,
                                                                     vertex_i)
                                    for vertex_j in pert_vertex_indices
                                    if (vertex_i, vertex_j, False)
                                    in normal_props)
        integral += "".join("e^{-\\tau_%i %s}"
                            % (vertex, self.vert_exp[vertex])
                            for vertex in pert_vertex_indices)
        return integral

    def attribute_qp_labels(self):
        """Attribute the appropriate qp labels to the graph's propagators."""
        idx_counter = 1
        for prop in self.graph.edges(keys=True, data=True):
            if prop[3]['anomalous']:
                prop[3]['qp_state'] = "k_{%i}k_{%i}" % (idx_counter + 1,
                                                        idx_counter)
                idx_counter += 2
            else:
                prop[3]['qp_state'] = "k_{%i}" % idx_counter
                idx_counter += 1

    def vertex_expression(self, vertex):
        """Return the expression associated to a given vertex.

        Args:
            vertex (int): The vertex of interest in the graph.

        Returns:
            (str): The LaTeX expression associated to the vertex.

        """
        expression = r"\epsilon^{" \
            + "".join("%s"
                      % prop[3]['qp_state']
                      for prop
                      in self.graph.out_edges(vertex, keys=True, data=True)
                      if not prop[3]['anomalous']) \
            + "}_{" \
            + "".join("%s"
                      % (prop[3]['qp_state'].split("}")[0] + "}"
                         if prop[3]['anomalous'] else prop[3]['qp_state'])
                      for prop
                      in self.graph.in_edges(vertex, keys=True, data=True)) \
            + "".join("%s"
                      % (prop[3]['qp_state'].split("}")[1] + "}")
                      for prop
                      in self.graph.out_edges(vertex, keys=True, data=True)
                      if prop[3]['anomalous']) \
            + "}"
        return expression

    def has_sign_factor(self):
        """Return True if a sign factor is associated to the diagram.

        Wrapper allowing for easy refactoring of expression code.

        Returns:
            (boolean): The presence of a sign factor.

        """
        # Use exclusive or for the sign factor
        return xor(self.has_crossing_sign(), self.has_anom_props_linked_sign())

    def extract_numerator(self):
        """Return the numerator associated to a PBMBPT graph.

        Returns:
            (str): The numerator of the graph.

        """
        graph = self.graph
        numerator = ""
        for vertex in graph:
            # Attribute the correct operator to each vertex
            numerator += r"\tilde{O}" if graph.nodes[vertex]['operator'] \
                else "\\Omega"
            # Attribute the good "type number" to each vertex
            numerator += "^{%i%i}_{" % (self.unsort_io_degrees[vertex][1],
                                        self.unsort_io_degrees[vertex][0])
            # First add the qp states corresponding to propagators going out
            numerator += "".join(prop[3]['qp_state']
                                 for prop in graph.out_edges(vertex,
                                                             keys=True,
                                                             data=True)
                                 if not prop[3]['anomalous'])

            # Add the qp states corresponding to propagators coming in
            previous_vertex = vertex - 1
            # Iterate over the lower vertices
            while previous_vertex >= 0:
                # Start with normal propagators
                numerator += "".join(
                    prop[3]['qp_state']
                    for prop in graph.in_edges(vertex, keys=True, data=True)
                    if prop[0] == previous_vertex and not prop[3]['anomalous'])
                # Read anomalous propagators backwards
                anom_props = [prop for prop
                              in graph.in_edges(vertex, keys=True, data=True)
                              if prop[0] == previous_vertex
                              and prop[3]['anomalous']]
                numerator += "".join(
                    (anom_prop[3]['qp_state'].split("}")[0] + "}")
                    for anom_prop in reversed(anom_props)
                    )
                previous_vertex -= 1
            # Add self-contractions
            numerator += "".join(
                (prop[3].split("}")[1] + "}" + prop[3].split("}")[0] + "}")
                for prop in graph.edges(vertex, keys=True, data='qp_state')
                if prop[0] == prop[1])
            # Add anomalous propagators entering from above
            numerator += "".join(prop[3]['qp_state'].split("}")[1] + "}"
                                 for prop in graph.out_edges(vertex,
                                                             keys=True,
                                                             data=True)
                                 if prop[3]['anomalous'] and prop[1] > prop[0])
            numerator += r"} (\varphi) " if graph.nodes[vertex]['operator'] \
                else "} "
        # Add the terms correspoding to anomalous propagators
        numerator += " ".join("R^{--}_{%s}(\\varphi)" % prop[3]['qp_state']
                              for prop
                              in graph.out_edges(keys=True, data=True)
                              if prop[3]['anomalous'])
        return numerator

    def symmetry_factor(self):
        """Return the overall symmetry factor of the diagram.

        Returns:
            (str): The combination of all symmetry factors.

        """
        sym_factor = ""
        vertex_sym = 1
        anom_factor = self.anomalous_contractions_factor()
        for vertex_degrees in self.unsort_io_degrees:
            if self.unsort_io_degrees.count(vertex_degrees) >= 2:
                vertex_sym = self.vertex_exchange_sym_factor
                break
        sym_factor += "%i" % (vertex_sym*anom_factor) \
            if (vertex_sym*anom_factor) > 1 else ""
        sym_factor += self.multiplicity_symmetry_factor()
        return sym_factor

    def multiplicity_symmetry_factor(self):
        """Return the symmetry factor associated with propagators multiplicity.

        Returns:
            (str): The symmetry factor associated with equivalent lines.

        """
        factor = ""
        # Account for up to three-body operators
        prop_multiplicity = [0 for _ in range(6)]
        for vertex_i in self.graph:
            for vertex_j in self.graph:
                nb_normal_props = sum(1 for prop
                                      in self.graph.edges(vertex_i,
                                                          keys=True,
                                                          data='anomalous')
                                      if prop[1] == vertex_j and not prop[3])
                nb_anomalous_props = sum(1 for prop
                                         in self.graph.edges(vertex_i,
                                                             keys=True,
                                                             data='anomalous')
                                         if prop[1] == vertex_j and prop[3])
                if nb_anomalous_props >= 2:
                    prop_multiplicity[nb_anomalous_props - 1] += 1
                if nb_normal_props >= 2:
                    prop_multiplicity[nb_normal_props - 1] += 1

        for prop_id, multiplicity in enumerate(prop_multiplicity):
            if multiplicity == 1:
                factor += "(%i!)" % (prop_id+1)
            elif multiplicity >= 2:
                factor += "(%i!)" % (prop_id+1) + "^%i" % multiplicity
        return factor

    def anomalous_contractions_factor(self):
        """Return the factor associated with anomalous self-contractions.

        Returns:
            (int): The anomalous self-contractions factor.

        """
        nb_self_contractions = len(list(nx.selfloop_edges(self.graph)))
        return pow(2, nb_self_contractions)

    def equivalent_permutations(self):
        """Return the permutations generating equivalent diagrams.

        Returns:
            (list): Vertices permutations as dictionnaries.

        """
        op_nm = nx.algorithms.isomorphism.categorical_node_match('operator',
                                                                 False)
        anom_em = nx.algorithms.isomorphism.categorical_multiedge_match(
            'anomalous', False)

        perm_vertices = [vertex for vertex, degrees
                         in enumerate(self.unsort_io_degrees)
                         if not self.graph.nodes[vertex]['operator']
                         and self.unsort_io_degrees.count(degrees) >= 2]
        permutations = []
        for perm in itertools.permutations(perm_vertices):
            permuted_graph = nx.relabel_nodes(self.check_graph,
                                              dict(list(zip(perm_vertices,
                                                            perm))),
                                              copy=True)
            # Check for a permutation that leaves the graph unchanged
            # (only way to keep the edge list of the same length)
            intersection = copy.deepcopy(self.check_graph)
            intersection.remove_edges_from(e for e in self.check_graph.edges()
                                           if e not in permuted_graph.edges())
            check = nx.algorithms.isomorphism.DiGraphMatcher(self.check_graph,
                                                             intersection,
                                                             node_match=op_nm,
                                                             edge_match=anom_em)
            if check.is_isomorphic():
                permutations.append(dict(list(zip(perm_vertices, perm))))
        return permutations

    def has_anom_props_linked_sign(self):
        """Return True if there is a minus sign associated to anom props.

        Anomalous propagators departing to higher vertices introduce a sign
        factor if a normal propagator is going to an even higher vertex, as it
        departs from the canonical representation used for numerator
        extraction.

        Returns:
            (bool): The presence of the sign factor.

        """
        counter = 0
        for vertex in self.graph:
            for prop_a in self.graph.out_edges(vertex,
                                               keys=True,
                                               data='anomalous'):
                # Check that prop_a is anomalous and not a self-contraction
                if prop_a[3] and prop_a[1] != prop_a[0]:
                    for prop_b in self.graph.out_edges(vertex,
                                                       keys=True,
                                                       data='anomalous'):
                        # Check for a normal prop going to a higher vertex
                        if prop_b[1] > prop_a[1] and not prop_b[3]:
                            counter += 1
        # True if overall factor is odd
        return counter % 2 == 1

    def has_anom_non_selfcontracted_props(self):
        """Return True if the diagram has anomalous propagators.

        Returns:
            (bool): The presence of anomalous propagators.

        """
        for prop in self.graph.edges(keys=True, data='anomalous'):
            if prop[3] and prop[1] != prop[0]:
                return True
        return False

    def write_diag_exps(self, latex_file, norder):
        """Write the expressions associated to a diagram in the LaTeX file.

        Args:
            latex_file (file): The LaTeX outputfile of the program.
            norder (int): The order in BMBPT formalism.

        """
        latex_file.write(
            "\\begin{align}\n\\text{PO}%i.%i.%i\n" % (norder,
                                                      (self.tags[0] + 1),
                                                      (self.tags[1] + 1))
            + "&= %s" % self.feynman_exp
            + r" \nonumber \\" + "\n"
            + "&= %s\\end{align}\n" % self.diag_exp)

    @property
    def check_graph(self):
        """Return a graph that can be used for topological equivalence checks.

        Lazy-initialized to reduce memory and CPU costs as this operation
        requires a deep copy.

        Returns:
            (NetworkX MultiDiGraph): The graph with doubled anomalous props.

        """
        if self._check_graph is None:
            self._check_graph = adg.diag.create_checkable_diagram(self.graph)
        return self._check_graph

    def write_graph(self, latex_file, directory, write_time):
        """Write the PBMBPT graph and its associated TSD to the LaTeX file.

        Args:
            latex_file (file): The LaTeX output file of the program.
            directory (str): The path to the result folder.
            write_time (bool): ``True`` if we want informations on the
                associated TSDs.

        """
        latex_file.write('\n\\begin{center}\n')
        adg.diag.draw_diagram(directory, latex_file,
                              "%i_%i" % (self.tags[0], self.tags[1]), 'diag')
        if write_time:
            latex_file.write(
                '\\hspace{10pt} $\\rightarrow$ \\hspace{10pt} T%i:'
                % (self.time_tag + 1))
            adg.diag.draw_diagram(directory, latex_file,
                                  str(self.time_tag), 'time')
        latex_file.write('\n\\end{center}\n\n')



"""Routines and class for Bogoliubov MBPT diagrams."""

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import zip
from builtins import range
from adg.tools import reversed_enumerate
import itertools
import numpy as np
import networkx as nx
import adg.tsd
import adg.diag


def diagrams_generation(p_order, three_body_use, nbody_obs, canonical):
    """Generate diagrams for BMBPT from bottom up.

    Args:
        p_order (int): The BMBPT perturbative order of the studied diagrams.
        three_body_use (bool): Flag for the use of three-body forces.
        nbody_obs (int): N-body character of the obervable of interest.
        canonical (bool): ``True`` if one draws only canonical diagrams.

    Returns:
        (list): NumPy arrays encoding the adjacency matrices of the graphs.

    >>> diags = diagrams_generation(1, False, 2, False)
    >>> len(diags)
    2
    >>> any(np.array_equal([[0, 4], [0, 0]], diag) for diag in diags)
    True
    >>> any(np.array_equal([[0, 2], [0, 0]], diag) for diag in diags)
    True
    >>> diags = diagrams_generation(1, True, 3, False)
    >>> len(diags)
    3
    >>> any(np.array_equal([[0, 6], [0, 0]], diag) for diag in diags)
    True
    >>> any(np.array_equal([[0, 4], [0, 0]], diag) for diag in diags)
    True
    >>> any(np.array_equal([[0, 2], [0, 0]], diag) for diag in diags)
    True
    >>> diags = diagrams_generation(2, False, 2, True)
    >>> len(diags)
    2
    >>> any(np.array_equal([[0, 2, 2], [0, 0, 2], [0, 0, 0]], d) for d in diags)
    True
    >>> any(np.array_equal([[0, 1, 1], [0, 0, 3], [0, 0, 0]], d) for d in diags)
    True

    """
    # Matrices contain operator vertex + p_order perturbative vertices
    order = p_order + 1

    # Create a null oriented adjacency matrix of dimension (p_order,p_order)
    matrices = [np.zeros((order, order), dtype=int)]

    # Generate oriented adjacency matrices going vertex-wise
    vertices = list(range(order))
    add = matrices.append
    for vertex in vertices:
        if vertex == 0:
            deg_max = 2*nbody_obs
        else:
            deg_max = 6 if three_body_use else 4
        for sum_index in range(vertex+1, order):
            for mat in reversed(matrices):
                elem_max = deg_max - sum(mat[k][vertex] + mat[vertex][k]
                                         for k in vertices)
                for elem in range(1, elem_max + 1):
                    temp_mat = mat.copy()
                    temp_mat[vertex, sum_index] = elem
                    add(temp_mat)
        adg.diag.check_vertex_degree(
            matrices, three_body_use, nbody_obs, canonical, vertex
        )
        if 0 < vertex < order-1:
            check_unconnected_spawn(matrices, vertex)
    remove_disconnected_matrices(matrices)
    return order_and_remove_topologically_equiv(matrices, order - 1)


def remove_disconnected_matrices(matrices):
    """Remove matrices corresponding to disconnected diagrams.

    Args:
        matrices (list): List of adjacency matrices.

    """
    vertices = list(range(matrices[0].shape[0]))
    permutations = [[0] + list(k)
                    for k in itertools.permutations(vertices[1:])]
    for idx, matrix in reversed_enumerate(matrices):
        for reordering in permutations:
            mat = matrix[:, reordering][reordering, :]
            for vert in vertices[1:]:
                # Check if the permuted matrix is block-diagonal
                if not mat[vertices[:vert], :][:, vertices[vert:]].any() \
                        and not mat[:, vertices[:vert]
                                    ][vertices[vert:], :].any():
                    del matrices[idx]
                    break
            else:
                continue
            break


def order_and_remove_topologically_equiv(matrices, max_vertex):
    """Order the matrices in sub-list and remove topologically equivalent ones.

    Args:
        matrices (list): The adjacency matrices to be checked.
        max_vertex (int): The maximum vertex which has been filled.

    Returns:
        (list): The ordered topologically unique matrices.

    """
    matrices_dict = {}
    for idx, matrix in reversed_enumerate(matrices):
        row0 = "".join("%i" % elem for elem
                       in np.sort(matrix[0, :]).tolist())
        if row0 in matrices_dict:
            matrices_dict[row0].append(matrix)
        else:
            matrices_dict[row0] = [matrix]
        del matrices[idx]
    for row_key in sorted(matrices_dict.keys()):
        matrices_dict[row_key] = \
            check_topologically_equivalent(matrices_dict[row_key], max_vertex)
    matrices = []
    for matrices_list in list(matrices_dict.values()):
        matrices += matrices_list
    return matrices


def check_topologically_equivalent(matrices, max_vertex):
    """Exclude matrices that would spawn topologically equivalent graphs.

    Args:
        matrices (list): Adjacency matrices to be checked.
        max_vertex (int): The maximum vertex which have been filled.

    Returns:
        (list): The topologically unique matrices.

    >>> import numpy
    >>> mats = [numpy.array([[0, 2, 0, 0], [2, 0, 0, 1], [0, 0, 0, 0], [0, 0, 0, 0]]), \
                numpy.array([[0, 0, 2, 0], [0, 0, 0, 0], [2, 0, 0, 1], [0, 0, 0, 0]])]
    >>>
    >>> mats = check_topologically_equivalent(mats, 2)
    >>> mats # doctest: +NORMALIZE_WHITESPACE
    [array([[0, 2, 0, 0], [2, 0, 0, 1], [0, 0, 0, 0], [0, 0, 0, 0]])]
    >>>
    >>> mats = check_topologically_equivalent([], 2)
    >>> mats # doctest: +NORMALIZE_WHITESPACE
    []

    """
    if not matrices:
        return []
    vertices = list(range(matrices[0].shape[0]))
    permutations = [[0] + list(k) + vertices[max_vertex+1:]
                    for k in itertools.permutations(vertices[1:max_vertex+1])]
    for ind_mat1, mat1 in reversed_enumerate(matrices[:-1]):
        mat1_1plus_sorted = np.sort(mat1[1:max_vertex+1, :].flat)
        for ind_mat2 in range(len(matrices)-1, ind_mat1, -1):
            mat2 = matrices[ind_mat2]
            # Basic check to avoid needless permutations
            if not (mat1_1plus_sorted
                    - np.sort(mat2[1:max_vertex+1, :].flat)).any():
                # Test for all possible permutations
                for reordering in permutations:
                    if not (mat1 - mat2[:, reordering][reordering, :]).any():
                        del matrices[ind_mat2]
                        break
                else:
                    continue
                break
    return matrices


def check_unconnected_spawn(matrices, max_filled_vertex):
    """Exclude some matrices that would spawn unconnected diagrams.

    Do several permutations among the rows and columns corresponding to already
    filled vertices, and check if one obtains a block-diagonal organisation,
    where the off-diagonals blocks connecting the already-filled and
    yet-unfilled parts of the matrix would be empty. In that case, remove the
    matrix.

    Args:
        matrices (list): The adjacency matrices to be checked.
        max_filled_vertex (int): The furthest vertex until which the matrices
            have been filled.

    >>> import numpy
    >>> mats = [numpy.array([[0, 2, 0], [2, 0, 0], [0, 0, 0]]), \
                numpy.array([[0, 2, 1], [2, 0, 1], [0, 0, 0]])]
    >>>
    >>> check_unconnected_spawn(mats, 1)
    >>> mats # doctest: +NORMALIZE_WHITESPACE
    [array([[0, 2, 1], [2, 0, 1], [0, 0, 0]])]

    """
    vertices = list(range(matrices[0].shape[0]))
    permutations = [[0] + list(k) + vertices[max_filled_vertex+1:]
                    for k
                    in itertools.permutations(vertices[1:max_filled_vertex+1])]
    for ind_mat, matrix in reversed_enumerate(matrices):
        # Test for all possible permutations
        for reordering in permutations:
            mat = matrix[:, reordering][reordering, :]
            # Check for non-zero elements in off-diagonal blocks
            if not mat[:, vertices[:max_filled_vertex+1]
                       ][vertices[max_filled_vertex+1:], :].any() \
                    and not mat[vertices[:max_filled_vertex+1],
                                :][:, vertices[max_filled_vertex+1:]].any():
                del matrices[ind_mat]
                break


def write_header(tex_file, commands, diags_nbs):
    """Write overall header for BMBPT result file.

    Args:
        tex_file (file): The ouput LaTeX file of the program.
        commands (Namespace): Flags for the program run.
        diags_nbs (dict): The number of diagrams per type.

    """
    tex_file.write(
        "Valid diagrams: %i\n\n" % diags_nbs['nb_diags']
        + "2N valid diagrams: %i\n\n" % diags_nbs['nb_2']
        + "2N canonical diagrams for the energy: %i\n\n" % diags_nbs['nb_2_hf']
        + "2N canonical diagrams for a generic operator only: %i\n\n"
        % diags_nbs['nb_2_ehf']
    )
    if not commands.canonical:
        tex_file.write(
            "2N non-canonical diagrams: %i\n\n" % diags_nbs['nb_2_not_hf']
        )
    if commands.with_3NF:
        tex_file.write(
            "3N valid diagrams: %i\n\n" % diags_nbs['nb_3']
            + "3N canonical diagrams for the energy: %i\n\n"
            % diags_nbs['nb_3_hf']
            + "3N canonical diagrams for a generic operator only: %i\n\n"
            % diags_nbs['nb_3_ehf']
        )
        if not commands.canonical:
            tex_file.write(
                "3N non-canonical diagrams: %i\n\n" % diags_nbs['nb_3_not_hf']
            )


def produce_expressions(diagrams, diagrams_time):
    """Produce and store the expressions associated to the BMBPT diagrams.

    Args:
        diagrams (list): The list of all BmbptFeynmanDiagrams.
        diagrams_time (list): Their associates TSDs.

    """
    for diag in diagrams:
        diag.attribute_qp_labels()
        for t_diag in diagrams_time:
            if diag.unique_id in t_diag.tags[1:]:
                diag.time_tag = t_diag.tags[0]
                diag.tsd_is_tree = t_diag.is_tree
                break
        diag.attribute_expressions(diagrams_time[diag.time_tag])


def order_diagrams(diagrams):
    """Order the BMBPT diagrams and return number of diags for each type.

    Args:
        diagrams (list): Possibly redundant BmbptFeynmanDiagrams.

    Returns:
        (tuple): First element is the list of topologically unique, ordered
            diagrams. Second element is a dict with the number of diagrams
            for each major type. Third element is a dict with the identifiers
            of diagrams starting each output file section.

    """
    diagrams_2_hf = []
    diagrams_2_ehf = []
    diagrams_2_not_hf = []
    diagrams_3_hf = []
    diagrams_3_ehf = []
    diagrams_3_not_hf = []

    for i_diag, diag in reversed_enumerate(diagrams):
        if diag.two_or_three_body == 2:
            if diag.hf_type == "HF":
                diagrams_2_hf.append(diag)
            elif diag.hf_type == "EHF":
                diagrams_2_ehf.append(diag)
            elif diag.hf_type == "noHF":
                diagrams_2_not_hf.append(diag)
        elif diag.two_or_three_body == 3:
            if diag.hf_type == "HF":
                diagrams_3_hf.append(diag)
            elif diag.hf_type == "EHF":
                diagrams_3_ehf.append(diag)
            elif diag.hf_type == "noHF":
                diagrams_3_not_hf.append(diag)
        del diagrams[i_diag]

    diagrams = diagrams_2_hf + diagrams_2_ehf + diagrams_2_not_hf \
        + diagrams_3_hf + diagrams_3_ehf + diagrams_3_not_hf

    diags_nb_per_type = {
        'nb_2_hf': len(diagrams_2_hf),
        'nb_2_ehf': len(diagrams_2_ehf),
        'nb_2_not_hf': len(diagrams_2_not_hf),
        'nb_3_hf': len(diagrams_3_hf),
        'nb_3_ehf': len(diagrams_3_ehf),
        'nb_3_not_hf': len(diagrams_3_not_hf),
        'nb_diags': len(diagrams),
        'nb_2': (len(diagrams_2_hf) + len(diagrams_2_ehf)
                 + len(diagrams_2_not_hf)),
        'nb_3': (len(diagrams_3_hf) + len(diagrams_3_ehf)
                 + len(diagrams_3_not_hf))
    }

    section_flags = {
        'two_body_hf': diagrams_2_hf[0].unique_id if diagrams_2_hf else -1,
        'two_body_ehf': diagrams_2_ehf[0].unique_id if diagrams_2_ehf else -1,
        'two_body_not_hf': diagrams_2_not_hf[0].unique_id
                           if diagrams_2_not_hf else -1,
        'three_body_hf': diagrams_3_hf[0].unique_id if diagrams_3_hf else -1,
        'three_body_ehf': diagrams_3_ehf[0].unique_id
                          if diagrams_3_ehf else -1,
        'three_body_not_hf': diagrams_3_not_hf[0].unique_id
                             if diagrams_3_not_hf else -1
    }

    return diagrams, diags_nb_per_type, section_flags


class BmbptFeynmanDiagram(adg.diag.Diagram):
    """Describes a BMBPT Feynman diagram with its related properties.

    Attributes:
        two_or_three_body (int): The 2 or 3-body characted of the vertices.
        time_tag (int): The tag number associated to the diagram's
            associated TSD.
        tsd_is_tree (bool): The tree or non-tree character of the
            associated TSD.
        feynman_exp (str): The Feynman expression associated to the diagram.
        diag_exp (str): The Goldstone expression associated to the diagram.
        vert_exp (list): The expression associated to the vertices.
        hf_type (str): The Hartree-Fock, non-Hartree-Fock or Hartree-Fock for
            the energy operator only character of the graph.
        unique_id (int): A unique number associated to the diagram.
        vertex_exchange_sym_factor (int): Lazy-initialized symmetry factor
            associated to the vertex exchange, stored to avoid being computed
            several times.

    """

    __slots__ = ('two_or_three_body', 'time_tag', 'tsd_is_tree', 'feynman_exp',
                 'diag_exp', 'vert_exp', 'hf_type', 'unique_id',
                 '_vert_exchange_sym_fact')

    def __init__(self, nx_graph, tag_num):
        """Generate a BMBPT diagrams using a NetworkX graph.

        Args:
            nx_graph (NetworkX MultiDiGraph): The graph of interest.
            tag_num (int): The tag number associated to the diagram.

        """
        adg.diag.Diagram.__init__(self, nx_graph)
        self.two_or_three_body = 3 if self.max_degree == 6 else 2
        self.tags = [tag_num]
        self.time_tag = -1
        self.tsd_is_tree = False
        self.feynman_exp = ""
        self.diag_exp = ""
        self.vert_exp = []
        if 2 not in self.degrees:
            self.hf_type = "HF"
        elif 2 not in self.unsort_degrees[1:]:
            self.hf_type = "EHF"
        else:
            self.hf_type = "noHF"
        self.unique_id = tag_num
        self._vert_exchange_sym_fact = None

    def attribute_expressions(self, time_diag):
        """Attribute the correct Feynman and Goldstone expressions.

        Args:
            time_diag (TimeStructureDiagram): The associated TSD.

        """
        self.vert_exp = [self.vertex_expression(vertex)
                         for vertex in self.graph]
        numerator = self.extract_numerator()
        denominator = self.time_tree_denominator(
            nx.relabel_nodes(time_diag.graph, time_diag.perms[self.unique_id])
        ) if self.tsd_is_tree else ""

        extra_factor = "" if self.tsd_is_tree \
            else "\\left[" \
            + " + ".join("\\frac{1}{%s}"
                         % self.time_tree_denominator(
                             nx.relabel_nodes(equi_t_graph,
                                              time_diag.perms[self.unique_id]))
                         for equi_t_graph in time_diag.equivalent_trees) \
            + " \\right]"

        # Determine the pre-factor
        prefactor = "(-1)^%i " % (len(self.graph) - 1)
        if self.has_sign_factor():
            prefactor = "-%s" % prefactor
        sym_fact = self.symmetry_factor()
        prefactor = "\\frac{%s}{%s}\\sum_{k_i}" % (prefactor, sym_fact) \
            if sym_fact != "" else "%s\\sum_{k_i}" % prefactor

        # Set the Feynman and Goldstone expressions
        self.feynman_exp = \
            "\\lim\\limits_{\\tau \\to \\infty}%s%s\\int_{0}^{\\tau}%s\n" \
            % (prefactor, numerator, self.extract_integral())
        self.diag_exp = \
            "%s\\frac{%s}{%s} %s\n" % (prefactor, numerator,
                                       denominator, extra_factor) \
            if denominator != "" \
            else "%s%s%s\n" % (prefactor, numerator, extra_factor)

    def has_sign_factor(self):
        """Return True if a sign factor is associated to the diagram.

        Wrapper allowing for easy refactoring of expression code.

        Returns:
            (boolean): The presence of a sign factor.

        """
        # Use exclusive or for the sign factor
        return self.has_crossing_sign()

    def symmetry_factor(self):
        """Return the overall symmetry factor of the diagram.

        Returns:
            (str): The combination of all symmetry factors.
        """
        sym_factor = ""
        for vertex_degrees in self.unsort_io_degrees:
            if self.unsort_io_degrees.count(vertex_degrees) >= 2:
                vertex_sym = self.vertex_exchange_sym_factor
                sym_factor += "%i" % vertex_sym if vertex_sym > 1 else ""
                break
        sym_factor += self.multiplicity_symmetry_factor()
        return sym_factor

    def vertex_expression(self, vertex):
        """Return the expression associated to a given vertex.

        Args:
            vertex (int): The vertex of interest in the graph.

        Returns:
            (str): The LaTeX expression associated to the vertex.

        """
        expression = r"\epsilon^{" \
            + "".join("%s"
                      % prop[3]['qp_state']
                      for prop
                      in self.graph.out_edges(vertex, keys=True, data=True)) \
            + "}_{" \
            + "".join("%s"
                      % prop[3]['qp_state']
                      for prop
                      in self.graph.in_edges(vertex, keys=True, data=True)) \
            + "}"
        return expression

    def write_graph(self, latex_file, directory, write_time):
        """Write the BMBPT graph and its associated TSD to the LaTeX file.

        Args:
            latex_file (file): The LaTeX output file of the program.
            directory (str): The path to the result folder.
            write_time (bool): ``True`` if we want informations on the
                associated TSDs.

        """
        latex_file.write('\n\\begin{center}\n')
        adg.diag.draw_diagram(directory, latex_file, str(self.tags[0]), 'diag')
        if write_time:
            latex_file.write(
                '\\hspace{10pt} $\\rightarrow$ \\hspace{10pt} T%i:'
                % (self.time_tag + 1))
            adg.diag.draw_diagram(directory, latex_file,
                                  str(self.time_tag), 'time')
        latex_file.write('\n\\end{center}\n\n')

    def write_tsd_info(self, diagrams_time, latex_file):
        """Write info related to the BMBPT associated TSD to the LaTeX file.

        Args:
            diagrams_time (list): The associated TSDs.
            latex_file (file): The LaTeX output file of the program.

        """
        for tdiag in diagrams_time:
            if self.time_tag == tdiag.tags[0]:
                time_diag = tdiag
                break
        latex_file.write(
            "\\begin{equation}\n\\text{T}%i = " % (self.time_tag + 1)
            + "%s\\end{equation}\n" % time_diag.expr)
        self.write_vertices_values(latex_file, time_diag.perms[self.unique_id])

    def write_section(self, result, commands, section_flags):
        """Write section and subsections for BMBPT result file.

        Args:
            result (file): The LaTeX output file of the program.
            commands (dict): The flags associated with run management.
            section_flags (dict): UniqueIDs of diags starting each section.

        """
        if self.unique_id == section_flags['two_body_hf']:
            result.write(
                "\\section{Two-body diagrams}\n\n"
                + "\\subsection{Two-body energy canonical diagrams}\n\n")
        elif self.unique_id == section_flags['two_body_ehf']:
            result.write("\\subsection{Two-body canonical diagrams " +
                         "for a generic operator only}\n\n")
        elif self.unique_id == section_flags['two_body_not_hf']:
            result.write("\\subsection{Two-body non-canonical diagrams}\n\n")
        if commands.with_3NF:
            if self.unique_id == section_flags['three_body_hf']:
                result.write(
                    "\\section{Three-body diagrams}\n\n"
                    + "\\subsection{Three-body energy canonical diagrams}\n\n")
            elif self.unique_id == section_flags['three_body_ehf']:
                result.write("\\subsection{Three-body canonical diagrams " +
                             "for a generic operator only}\n\n")
            elif self.unique_id == section_flags['three_body_not_hf']:
                result.write(
                    "\\subsection{Three-body non-canonical diagrams}\n\n")
        if commands.theory == "PBMBPT":
            result.write("\\paragraph{Diagram %i.%i:}\n" % (self.tags[0] + 1,
                                                            self.tags[1] + 1))
        else:
            result.write("\\paragraph{Diagram %i:}\n" % (self.tags[0] + 1))
        self.write_diag_exps(result, commands.order)

    def write_vertices_values(self, latex_file, mapping):
        """Write the qp energies associated to each vertex of the diag.

        Args:
            latex_file (file): The LaTeX output file of the program.
            mapping (dict): A mapping between the vertices in the diagram and
                the vertices in its euivalent TSD, since permutations between
                vertices are possible.

        """
        latex_file.write("\\begin{align*}\n")
        for ind in range(1, len(self.vert_exp)):
            latex_file.write("a_%i &= %s" % (ind, self.vert_exp[mapping[ind]]))
            if ind != len(self.vert_exp)-1:
                latex_file.write(r"\\")
            latex_file.write('\n')
        latex_file.write("\\end{align*}\n")

    def write_diag_exps(self, latex_file, norder):
        """Write the expressions associated to a diagram in the LaTeX file.

        Args:
            latex_file (file): The LaTeX outputfile of the program.
            norder (int): The order in BMBPT formalism.

        """
        latex_file.write(
            "\\begin{align}\n\\text{PO}%i.%i\n" % (norder, (self.tags[0] + 1))
            + "&= %s" % self.feynman_exp
            + r" \nonumber \\" + "\n"
            + "&= %s\\end{align}\n" % self.diag_exp)

    @property
    def vertex_exchange_sym_factor(self):
        """Return the symmetry factor associated with vertex exchange.

        Returns:
            (int): The symmetry factor for vertex exchange.

        """
        if self._vert_exchange_sym_fact is None:
            factor = len(self.equivalent_permutations())
            self._vert_exchange_sym_fact = max(factor, 1)
        return self._vert_exchange_sym_fact

    def equivalent_permutations(self):
        """Return the permutations generating equivalent diagrams.

        Returns:
            (list): Vertices permutations as dictionnaries.

        """
        perm_vertices = [vertex for vertex, degrees
                         in enumerate(self.unsort_io_degrees)
                         if not self.graph.nodes[vertex]['operator']
                         and self.unsort_io_degrees.count(degrees) >= 2]
        permutations = []
        op_nm = nx.algorithms.isomorphism.categorical_node_match('operator',
                                                                 False)
        for permutation in itertools.permutations(perm_vertices):
            permuted_graph = nx.relabel_nodes(self.graph,
                                              dict(list(zip(perm_vertices,
                                                            permutation))),
                                              copy=True)
            inter = nx.intersection(self.graph, permuted_graph)
            inter.nodes[0]['operator'] = True
            # Check for a permutation that leaves the graph unchanged
            # (only way to keep the edge list of the same length)
            # (is_isomorphic could maybe be replaced by something faster here)
            if nx.is_isomorphic(self.graph, inter, node_match=op_nm):
                permutations.append(dict(list(zip(perm_vertices,
                                                  permutation))))
        return permutations

    def extract_integral(self):
        """Return the integral part of the Feynman expression of the diag.

        Returns:
            (str): The integral part of its Feynman expression.

        """
        pert_vertex_indices = list(range(1, len(self.graph)))
        integral = "".join("\\mathrm{d}\\tau_%i" % vertex
                           for vertex in pert_vertex_indices)
        if len(pert_vertex_indices) > 1:
            for vertex_i in pert_vertex_indices:
                integral += "".join("\\theta(\\tau_%i-\\tau_%i) " % (vertex_j,
                                                                     vertex_i)
                                    for vertex_j in pert_vertex_indices
                                    if self.graph.has_edge(vertex_i, vertex_j))
        integral += "".join("e^{-\\tau_%i %s}"
                            % (vertex, self.vert_exp[vertex])
                            for vertex in pert_vertex_indices)
        return integral

    def attribute_qp_labels(self):
        """Attribute the appropriate qp labels to the graph's propagators."""
        for idx, prop in enumerate(self.graph.edges(keys=True, data=True)):
            prop[3]['qp_state'] = "k_{%i}" % (idx+1)

    def extract_numerator(self):
        """Return the numerator associated to a BMBPT graph.

        Returns:
            (str): The numerator of the graph.

        """
        graph = self.graph
        numerator = ""
        for vertex in graph:
            # Attribute the correct operator to each vertex
            numerator += "O" if graph.nodes[vertex]['operator'] else "\\Omega"
            # Attribute the good "type number" to each vertex
            numerator += "^{%i%i}_{" % (self.unsort_io_degrees[vertex][1],
                                        self.unsort_io_degrees[vertex][0])
            # First add the qp states corresponding to propagators going out
            numerator += "".join(prop[3]['qp_state']
                                 for prop
                                 in graph.out_edges(vertex,
                                                    keys=True, data=True))
            # Add the qp states corresponding to propagators coming in
            for previous_vertex in reversed(range(vertex)):
                numerator += "".join(
                    prop[3]['qp_state']
                    for prop in graph.in_edges(vertex, keys=True, data=True)
                    if prop[0] == previous_vertex)
            numerator += "} "
        return numerator

    def has_crossing_sign(self):
        """Return True for a minus sign associated with crossing propagators.

        Use the fact that all lines propagate upwards and the
        canonical representation of the diagrams and vertices.

        Returns:
            (bool): Encode for the sign factor associated with crossing
                propagators.

        """
        nb_crossings = 0
        for vertex in self.graph:
            for propagator in self.graph.out_edges(vertex, keys=True):
                for vertex_ante in range(propagator[0]):
                    for vertex_post in range(propagator[0]+1, propagator[1]):
                        nb_crossings += self.graph.number_of_edges(vertex_ante,
                                                                   vertex_post)
        return nb_crossings % 2 == 1

    def multiplicity_symmetry_factor(self):
        """Return the symmetry factor associated with propagators multiplicity.

        Returns:
            (str): The symmetry factor associated with equivalent lines.

        """
        factor = ""
        # Account for up to three-body operators
        prop_multiplicity = [0 for _ in range(6)]
        for vertex_i in self.graph:
            for vertex_j in self.graph:
                if self.graph.number_of_edges(vertex_i, vertex_j) >= 2:
                    prop_multiplicity[self.graph.number_of_edges(
                        vertex_i, vertex_j) - 1] += 1

        for prop_id, multiplicity in enumerate(prop_multiplicity):
            if multiplicity == 1:
                factor += "(%i!)" % (prop_id+1)
            elif multiplicity >= 2:
                factor += "(%i!)" % (prop_id+1) + "^%i" % multiplicity
        return factor

    def time_tree_denominator(self, time_graph):
        """Return the denominator for a time-tree graph.

        Args:
            time_graph (NetworkX MultiDiGraph): Its associated time-structure
                graph.

        Returns:
            (str): The denominator of the graph.

        """
        denominator = ""
        for vertex_i in range(1, len(time_graph)):
            subgraph_stack = [vertex_j for vertex_j
                              in nx.descendants(time_graph, vertex_i)]
            subgraph_stack.append(vertex_i)
            subdiag = self.graph.subgraph(subgraph_stack)
            denominator += "%s\\ " % adg.diag.extract_denom(self.graph,
                                                            subdiag)
        return denominator



"""Miscellaneous diagram-unrelated tools for ADG."""


from builtins import object, zip
from itertools import count


def reversed_enumerate(data):
    """Return the index and item of the data in reversed order.

    Args:
        data (iterable data structure): The data to be used..

    Returns:
        (tuple): Index and item.

    >>> list(reversed_enumerate(['A', 'B', 'C']))
    [(2, 'C'), (1, 'B'), (0, 'A')]

    """
    for index, item in zip(count(len(data) - 1, -1), reversed(data)):
        yield index, item


class UniqueID(object):
    """Counter making sure of generating a unique ID number for diagrams.

    Attributes:
        current (int): The unique identifier to be attributedto a diagram.

    """

    __slots__ = 'current'

    def __init__(self):
        """Initialize counter with value 0."""
        self.current = 0

    def get(self):
        """Iterate on counter value and return current value.

        Returns:
            (int): A unique identifier for the diagram.

        """
        self.current += 1
        return self.current - 1



"""Main routine of the Automated Diagram Generator."""
from __future__ import print_function

from datetime import datetime
import sys
import adg.run
import adg.bmbpt
import adg.bimsrg
import adg.mbpt
import adg.tools
import adg.tsd


def main():
    """Launch the ADG program."""
    print(
        "#####################\n"
        + "# Automatic Diagram #\n"
        + "#     Generator     #\n"
        + "#      v%s       #\n" % adg.__version__
        + "#                   #\n"
        + "#  by ADG Dev Team  #\n"
        + "#####################\n"
    )
    run_commands = adg.run.parse_command_line(sys.argv[1:])

    if run_commands.interactive:
        run_commands = adg.run.interactive_interface(run_commands)

    directory = adg.run.attribute_directory(run_commands)

    # Start computing everything
    print("Running")
    start_time = datetime.now()

    id_gene = adg.tools.UniqueID()

    diagrams = adg.run.generate_diagrams(run_commands, id_gene)

    # Ordering the diagrams in a convenient way
    diagrams, diags_nbs, section_flags = adg.run.order_diagrams(diagrams,
                                                                run_commands)

    # Produce TSD for the expressions of BMBPT diagrams
    if run_commands.theory in ("BMBPT", "PBMBPT"):

        diagrams_time = [adg.tsd.TimeStructureDiagram(diagram)
                         for diagram in diagrams]

        diagrams_time, nb_tree_tsds = adg.tsd.treat_tsds(diagrams_time)

        adg.bmbpt.produce_expressions(diagrams, diagrams_time)

    else:
        diagrams_time = []

    print("Time elapsed: ", datetime.now() - start_time)

    adg.run.print_diags_numbers(run_commands, diags_nbs)

    # Writing a feynmp file for each graph
    if run_commands.draw_diags:
        adg.run.prepare_drawing_instructions(directory, run_commands,
                                             diagrams, diagrams_time)

    # Write everything down in a nice LaTeX file
    latex_file = open(directory + '/result.tex', 'w')

    adg.run.write_file_header(latex_file, run_commands, diags_nbs)

    if run_commands.theory in ("BMBPT", "PBMBPT") and run_commands.draw_tsds:
        adg.tsd.write_section(latex_file, directory, run_commands.draw_diags,
                              diagrams_time, nb_tree_tsds, diagrams)
    for diag in diagrams:
        diag.write_section(latex_file, run_commands, section_flags)

        if run_commands.draw_diags:
            diag.write_graph(latex_file, directory, run_commands.draw_tsds)

        if run_commands.theory in ("BMBPT", "PBMBPT") \
                and run_commands.draw_tsds:
            diag.write_tsd_info(diagrams_time, latex_file)

    latex_file.write("\\end{document}")
    latex_file.close()

    adg.diag.print_adj_matrices(directory, diagrams)

    # Produce an output adapted to Christian Drischler's format
    if run_commands.cd_output:
        adg.mbpt.print_cd_output(directory, diagrams)

    if run_commands.compile:
        adg.run.compile_manager(directory)

    adg.run.clean_folders(directory, run_commands)

    print("\nResult saved in ./%s/" % directory)
    print("\nADG ended successfully!\n")


if __name__ == "__main__":
    main()
